"""
Copyright 2015-2024 Jean-Baptiste Delisle

This file is part of BEd.

BEd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BEd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with BEd.  If not, see <http://www.gnu.org/licenses/>.
"""

# This the main file of BEd
# It creates the main window
# and all the user-callable functions (using the menu, shortcuts...)

# Important vocabulary conventions:
# When compiling a Beamer presentation in pdf,
# a single Beamer "frame" can be spread over several pdf "pages",
# if the frame contains some overlay commands (\only...).
# "frame" corresponds to the Beamer definition
# "page" refers to the pages in the final pdf
# "framePage" corresponds to the page number in a given Beamer frame

import html
import os
import re
import sys
import tempfile
import traceback

from PySide6 import QtCore, QtGui, QtWidgets

from . import parsing
from .document import Document
from .element import Arrow, Image, Text, TikzPicture
from .frame import Frame
from .group import Group
from .helpers import striplines
from .painter import PainterArea
from .settings import settings
from .undo import UndoStack

_exec_path = os.path.dirname(os.path.realpath(__file__))
_icons_path = f'{_exec_path}/icons'
_tr_path = f'{_exec_path}/translation'
_latex_path = f'{_exec_path}/latex'


class BEd(QtWidgets.QMainWindow):
  def __init__(self):
    """Initialisation of application"""
    super().__init__()
    self.compileOutput = tempfile.NamedTemporaryFile(prefix='bed-', suffix='.log').name
    self.document = None
    self.tmpFile = None
    self.frame = None
    self.framePage = None
    self.isCompiling = False
    self.copyList = []
    # Init of the window properties...
    self.setMinimumSize(550, 550)
    # set initial size of window (from config file)
    self.resize(settings.window_width, settings.window_height)
    self.setWindowTitle('BEd')
    self.setWindowIcon(QtGui.QIcon(f'{_icons_path}/bed.svg'))
    # Init of the area where presentation slides will be drawn (painterArea)
    self.painterArea = PainterArea(self)
    self.setCentralWidget(self.painterArea.scrollArea)
    self.painterArea.setFocus()
    # Init of actions, menus, pagebar
    self.initMenus()
    self.show()

  def initDocument(self, filename=None, page=0):
    """Initialisation of document"""
    if filename is not None:
      # If filename is passed in arg open the file
      self.openFile(filename, page)
      # Init of undo/redo stack
      self.undoStack = UndoStack(self)
    else:
      # Else start a new document from scratch
      self.newDoc()
      # Go to home directory
      os.chdir(os.path.expanduser('~'))

  def initMenus(self):
    """Initialisation of menus..."""
    # Status bar
    self.statusBar()

    # Cleaning previously defined actions
    # Needed to avoid shortcuts ambiguity when doing a resetUI
    for action in self.actions():
      self.removeAction(action)

    # Actions and shortcuts (from config file)
    newDocAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'document-new', QtGui.QIcon(f'{_icons_path}/document-new.svg')
      ),
      self.tr('&New document'),
      self,
    )
    newDocAction.setShortcut(QtGui.QKeySequence.New)
    newDocAction.setStatusTip(self.tr('Start a new document', 'BEd'))
    newDocAction.triggered.connect(self.newDoc)
    self.addAction(newDocAction)

    openAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'document-open', QtGui.QIcon(f'{_icons_path}/document-open.svg')
      ),
      self.tr('&Open'),
      self,
    )
    openAction.setShortcut(QtGui.QKeySequence.Open)
    openAction.setStatusTip(self.tr('Open TeX file'))
    openAction.triggered.connect(self.openDialog)
    self.addAction(openAction)

    reloadAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'view-refresh', QtGui.QIcon(f'{_icons_path}/view-refresh.svg')
      ),
      self.tr('&Reload'),
      self,
    )
    reloadAction.setShortcut(QtGui.QKeySequence(settings.shortcut_reload))
    reloadAction.setStatusTip(self.tr('Reload TeX file'))
    reloadAction.triggered.connect(self.reloadTeX)
    self.addAction(reloadAction)

    previewAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'view-preview', QtGui.QIcon(f'{_icons_path}/view-preview.svg')
      ),
      self.tr('&Preview'),
      self,
    )
    previewAction.setShortcut(QtGui.QKeySequence(settings.shortcut_preview))
    previewAction.setStatusTip(self.tr('Preview current frame'))
    previewAction.triggered.connect(self.previewCurrentFrame)
    self.addAction(previewAction)

    saveAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'document-save', QtGui.QIcon(f'{_icons_path}/document-save.svg')
      ),
      self.tr('&Save'),
      self,
    )
    saveAction.setShortcut(QtGui.QKeySequence.Save)
    if settings.auto_reload:
      saveAction.setStatusTip(self.tr('Save TeX file, compile, and reload'))
    else:
      saveAction.setStatusTip(self.tr('Save TeX file'))
    saveAction.triggered.connect(self.save)
    self.addAction(saveAction)

    saveAsAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'document-save-as', QtGui.QIcon(f'{_icons_path}/document-save-as.svg')
      ),
      self.tr('Save &as...'),
      self,
    )
    saveAsAction.setShortcut(QtGui.QKeySequence.SaveAs)
    if settings.auto_reload:
      saveAsAction.setStatusTip(self.tr('Save TeX file as..., compile, and reload'))
    else:
      saveAsAction.setStatusTip(self.tr('Save TeX file as...'))
    saveAsAction.triggered.connect(self.saveAs)
    self.addAction(saveAsAction)

    exitAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'application-exit', QtGui.QIcon(f'{_icons_path}/application-exit.svg')
      ),
      self.tr('&Exit'),
      self,
    )
    exitAction.setShortcut(QtGui.QKeySequence.Quit)
    exitAction.setStatusTip(self.tr('Exit application'))
    exitAction.triggered.connect(self.close)
    self.addAction(exitAction)

    previousPageAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'go-previous', QtGui.QIcon(f'{_icons_path}/go-previous.svg')
      ),
      self.tr('Previous page'),
      self,
    )
    previousPageAction.setShortcut(QtGui.QKeySequence.MoveToPreviousChar)
    previousPageAction.setStatusTip(self.tr('Go to the previous page'))
    previousPageAction.triggered.connect(self.previousPage)
    self.addAction(previousPageAction)

    nextPageAction = QtGui.QAction(
      QtGui.QIcon.fromTheme('go-next', QtGui.QIcon(f'{_icons_path}/go-next.svg')),
      self.tr('Next page'),
      self,
    )
    nextPageAction.setShortcut(QtGui.QKeySequence.MoveToNextChar)
    nextPageAction.setStatusTip(self.tr('Go to the next page'))
    nextPageAction.triggered.connect(self.nextPage)
    self.addAction(nextPageAction)

    selectAllAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'edit-select-all', QtGui.QIcon(f'{_icons_path}/edit-select-all.svg')
      ),
      self.tr('Select &all...'),
      self,
    )
    selectAllAction.setShortcut(QtGui.QKeySequence.SelectAll)
    selectAllAction.setStatusTip(self.tr('Select all objects in current frame'))
    selectAllAction.triggered.connect(self.painterArea.selectAll)
    self.addAction(selectAllAction)

    unselectAllAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'dialog-cancel', QtGui.QIcon(f'{_icons_path}/dialog-cancel.svg')
      ),
      self.tr('Unselect &all...'),
      self,
    )
    unselectAllAction.setShortcut(QtGui.QKeySequence.Deselect)
    unselectAllAction.setStatusTip(self.tr('Unselect all objects in current frame'))
    unselectAllAction.triggered.connect(self.painterArea.unselectAll)
    self.addAction(unselectAllAction)

    editAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'document-edit', QtGui.QIcon(f'{_icons_path}/document-edit.svg')
      ),
      self.tr('&Edit...'),
      self,
    )
    editAction.setShortcut(QtGui.QKeySequence(settings.shortcut_edit))
    editAction.setStatusTip(self.tr('Edit selected objects'))
    editAction.triggered.connect(self.editAnyConnect)
    self.addAction(editAction)

    editDocAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'story-editor', QtGui.QIcon(f'{_icons_path}/story-editor.svg')
      ),
      self.tr('Edit &document'),
      self,
    )
    editDocAction.setShortcut(QtGui.QKeySequence(settings.shortcut_editDoc))
    editDocAction.setStatusTip(self.tr('Edit document parameters (LaTeX header...)'))
    editDocAction.triggered.connect(self.editDoc)
    self.addAction(editDocAction)

    copyAction = QtGui.QAction(
      QtGui.QIcon.fromTheme('edit-copy', QtGui.QIcon(f'{_icons_path}/edit-copy.svg')),
      self.tr('&Copy'),
      self,
    )
    copyAction.setShortcut(QtGui.QKeySequence.Copy)
    copyAction.setStatusTip(self.tr('Copy selected objects'))
    copyAction.triggered.connect(self.copyAny)
    self.addAction(copyAction)

    cutAction = QtGui.QAction(
      QtGui.QIcon.fromTheme('edit-cut', QtGui.QIcon(f'{_icons_path}/edit-cut.svg')),
      self.tr('&Cut'),
      self,
    )
    cutAction.setShortcut(QtGui.QKeySequence.Cut)
    cutAction.setStatusTip(self.tr('Cut selected objects'))
    cutAction.triggered.connect(self.cutAny)
    self.addAction(cutAction)

    pasteAction = QtGui.QAction(
      QtGui.QIcon.fromTheme('edit-paste', QtGui.QIcon(f'{_icons_path}/edit-paste.svg')),
      self.tr('&Paste'),
      self,
    )
    pasteAction.setShortcut(QtGui.QKeySequence.Paste)
    pasteAction.setStatusTip(self.tr('Paste selected objects'))
    pasteAction.triggered.connect(self.pasteAny)
    self.addAction(pasteAction)

    deleteAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'edit-delete', QtGui.QIcon(f'{_icons_path}/edit-delete.svg')
      ),
      self.tr('&Delete'),
      self,
    )
    deleteAction.setShortcut(QtGui.QKeySequence.Delete)
    deleteAction.setStatusTip(self.tr('Delete selected objects'))
    deleteAction.triggered.connect(self.deleteAny)
    self.addAction(deleteAction)

    undoAction = QtGui.QAction(
      QtGui.QIcon.fromTheme('edit-undo', QtGui.QIcon(f'{_icons_path}/edit-undo.svg')),
      self.tr('&Undo'),
      self,
    )
    undoAction.setShortcut(QtGui.QKeySequence.Undo)
    undoAction.setStatusTip(self.tr('Undo last action'))
    undoAction.triggered.connect(self.undo)
    self.addAction(undoAction)

    redoAction = QtGui.QAction(
      QtGui.QIcon.fromTheme('edit-redo', QtGui.QIcon(f'{_icons_path}/edit-redo.svg')),
      self.tr('&Redo'),
      self,
    )
    redoAction.setShortcut(QtGui.QKeySequence.Redo)
    redoAction.setStatusTip(self.tr('Redo action'))
    redoAction.triggered.connect(self.redo)
    self.addAction(redoAction)

    groupAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'object-group', QtGui.QIcon(f'{_icons_path}/object-group.svg')
      ),
      self.tr('&Group'),
      self,
    )
    groupAction.setShortcut(QtGui.QKeySequence(settings.shortcut_group))
    groupAction.setStatusTip(self.tr('Group selected objects'))
    groupAction.triggered.connect(self.group)
    self.addAction(groupAction)

    ungroupAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'object-ungroup', QtGui.QIcon(f'{_icons_path}/object-ungroup.svg')
      ),
      self.tr('&Ungroup'),
      self,
    )
    ungroupAction.setShortcut(QtGui.QKeySequence(settings.shortcut_ungroup))
    ungroupAction.setStatusTip(self.tr('Ungroup selected objects'))
    ungroupAction.triggered.connect(self.ungroup)
    self.addAction(ungroupAction)

    newFrameAction = QtGui.QAction(
      QtGui.QIcon.fromTheme('window-new', QtGui.QIcon(f'{_icons_path}/window-new.svg')),
      self.tr('New &frame'),
      self,
    )
    newFrameAction.setShortcut(QtGui.QKeySequence(settings.shortcut_newFrame))
    newFrameAction.setStatusTip(self.tr('Add a new frame'))
    newFrameAction.triggered.connect(self.newFrame)
    self.addAction(newFrameAction)

    newFrameTemplateActions = []
    for ktemp in range(len(settings.templates) // 2):
      newFrameTemplateActions.append(
        QtGui.QAction(
          QtGui.QIcon.fromTheme(
            'project-development-new-template',
            QtGui.QIcon(f'{_icons_path}/project-development-new-template.svg'),
          ),
          self.tr('Template') + f' &{ktemp + 1} "{settings.templates[2 * ktemp]}"',
          self,
        )
      )
      newFrameTemplateActions[-1].setShortcut(
        QtGui.QKeySequence(settings.shortcut_newFrameTemplate + f'+{ktemp + 1}')
      )
      newFrameTemplateActions[-1].setStatusTip(
        self.tr('Add a new frame from the template')
        + f' "{settings.templates[2 * ktemp]}"'
      )
      newFrameTemplateActions[-1].triggered.connect(
        lambda _, who=ktemp: self.newFrameTemplate(who)
      )
      self.addAction(newFrameTemplateActions[-1])

    saveTemplateAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'document-save-as-template',
        QtGui.QIcon(f'{_icons_path}/document-save-as-template.svg'),
      ),
      self.tr('Save &template'),
      self,
    )
    saveTemplateAction.setShortcut(QtGui.QKeySequence(settings.shortcut_saveTemplate))
    saveTemplateAction.setStatusTip(self.tr('Save current frame as template'))
    saveTemplateAction.triggered.connect(self.saveTemplate)
    self.addAction(saveTemplateAction)

    deleteTemplateAction = QtGui.QAction(
      QtGui.QIcon(f'{_icons_path}/project-development-delete-template.svg'),
      self.tr('&Delete template'),
      self,
    )
    deleteTemplateAction.setShortcut(
      QtGui.QKeySequence(settings.shortcut_deleteTemplate)
    )
    deleteTemplateAction.setStatusTip(self.tr('Delete one of the saved templates'))
    deleteTemplateAction.triggered.connect(self.deleteTemplate)
    self.addAction(deleteTemplateAction)

    newTextAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'insert-text-frame', QtGui.QIcon(f'{_icons_path}/insert-text-frame.svg')
      ),
      self.tr('New &text'),
      self,
    )
    newTextAction.setShortcut(QtGui.QKeySequence(settings.shortcut_newText))
    newTextAction.setStatusTip(self.tr('Add a new text block'))
    newTextAction.triggered.connect(self.newText)
    self.addAction(newTextAction)

    newImageAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'insert-image', QtGui.QIcon(f'{_icons_path}/insert-image.svg')
      ),
      self.tr('New &image'),
      self,
    )
    newImageAction.setShortcut(QtGui.QKeySequence(settings.shortcut_newImage))
    newImageAction.setStatusTip(self.tr('Add a new image'))
    newImageAction.triggered.connect(self.newImage)
    self.addAction(newImageAction)

    newTikzPictureAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'draw-triangle', QtGui.QIcon(f'{_icons_path}/draw-triangle.svg')
      ),
      self.tr('New &Tikz picture'),
      self,
    )
    newTikzPictureAction.setShortcut(
      QtGui.QKeySequence(settings.shortcut_newTikzPicture)
    )
    newTikzPictureAction.setStatusTip(self.tr('Add a new Tikz picture'))
    newTikzPictureAction.triggered.connect(self.newTikzPicture)
    self.addAction(newTikzPictureAction)

    newArrowAction = QtGui.QAction(
      QtGui.QIcon.fromTheme('draw-line', QtGui.QIcon(f'{_icons_path}/draw-line.svg')),
      self.tr('New &arrow'),
      self,
    )
    newArrowAction.setShortcut(QtGui.QKeySequence(settings.shortcut_newArrow))
    newArrowAction.setStatusTip(self.tr('Add a new arrow'))
    newArrowAction.triggered.connect(self.newArrow)
    self.addAction(newArrowAction)

    moveUpAction = QtGui.QAction(
      QtGui.QIcon.fromTheme('go-up', QtGui.QIcon(f'{_icons_path}/go-up.svg')),
      self.tr('Move &up'),
      self,
    )
    moveUpAction.setShortcut(QtGui.QKeySequence(settings.shortcut_moveUp))
    moveUpAction.setStatusTip(self.tr('Move up selection a notch'))
    moveUpAction.triggered.connect(self.moveUp)
    self.addAction(moveUpAction)

    moveDownAction = QtGui.QAction(
      QtGui.QIcon.fromTheme('go-down', QtGui.QIcon(f'{_icons_path}/go-down.svg')),
      self.tr('Move &down'),
      self,
    )
    moveDownAction.setShortcut(QtGui.QKeySequence(settings.shortcut_moveDown))
    moveDownAction.setStatusTip(self.tr('Move down selection a notch'))
    moveDownAction.triggered.connect(self.moveDown)
    self.addAction(moveDownAction)

    moveTopAction = QtGui.QAction(
      QtGui.QIcon.fromTheme('go-top', QtGui.QIcon(f'{_icons_path}/go-top.svg')),
      self.tr('Move &top'),
      self,
    )
    moveTopAction.setShortcut(QtGui.QKeySequence(settings.shortcut_moveTop))
    moveTopAction.setStatusTip(self.tr('Move selection to top level'))
    moveTopAction.triggered.connect(self.moveTop)
    self.addAction(moveTopAction)

    moveBottomAction = QtGui.QAction(
      QtGui.QIcon.fromTheme('go-bottom', QtGui.QIcon(f'{_icons_path}/go-bottom.svg')),
      self.tr('Move &bottom'),
      self,
    )
    moveBottomAction.setShortcut(QtGui.QKeySequence(settings.shortcut_moveBottom))
    moveBottomAction.setStatusTip(self.tr('Move selection to bottom level'))
    moveBottomAction.triggered.connect(self.moveBottom)
    self.addAction(moveBottomAction)

    self.activateGridAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'grid-rectangular', QtGui.QIcon(f'{_icons_path}/grid-rectangular.svg')
      ),
      self.tr('&Grid'),
      self,
      checkable=True,
    )
    self.activateGridAction.setShortcut(
      QtGui.QKeySequence(settings.shortcut_activateGrid)
    )
    self.activateGridAction.setStatusTip(self.tr('(De)activate grid guides'))
    self.activateGridAction.setChecked(settings.activate_grid)
    self.activateGridAction.triggered.connect(self.toggleGrid)
    self.addAction(self.activateGridAction)

    self.activateObjGuidesAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'format-border-set-all', QtGui.QIcon(f'{_icons_path}/format-border-set-all.svg')
      ),
      self.tr('&Object guides'),
      self,
      checkable=True,
    )
    self.activateObjGuidesAction.setShortcut(
      QtGui.QKeySequence(settings.shortcut_activateObjGuides)
    )
    self.activateObjGuidesAction.setStatusTip(self.tr('(De)activate object guides'))
    self.activateObjGuidesAction.setChecked(settings.activate_object_guides)
    self.activateObjGuidesAction.triggered.connect(self.toggleObjGuides)
    self.addAction(self.activateObjGuidesAction)

    self.showHiddenObjectsAction = QtGui.QAction(
      QtGui.QIcon.fromTheme('visibility', QtGui.QIcon(f'{_icons_path}/visibility.svg')),
      self.tr('&Hidden objects'),
      self,
      checkable=True,
    )
    self.showHiddenObjectsAction.setShortcut(
      QtGui.QKeySequence(settings.shortcut_showHiddenObjects)
    )
    self.showHiddenObjectsAction.setStatusTip(self.tr('Reveal hidden objects'))
    self.showHiddenObjectsAction.setChecked(settings.show_hidden_objects)
    self.showHiddenObjectsAction.triggered.connect(self.toggleHiddenObjects)
    self.addAction(self.showHiddenObjectsAction)

    self.hideMenuAction = QtGui.QAction(
      QtGui.QIcon.fromTheme('show-menu', QtGui.QIcon(f'{_icons_path}/show-menu.svg')),
      self.tr('Show/Hide &menu'),
      self,
      checkable=True,
    )
    self.hideMenuAction.setShortcut(QtGui.QKeySequence(settings.shortcut_hideMenu))
    self.hideMenuAction.setStatusTip(self.tr('Show/Hide menu bar'))
    self.hideMenuAction.setChecked(settings.menuVisible)
    self.hideMenuAction.triggered.connect(self.toggleMenu)
    self.addAction(self.hideMenuAction)

    editSettingsAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'document-edit', QtGui.QIcon(f'{_icons_path}/document-edit.svg')
      ),
      self.tr('&Edit settings'),
      self,
    )
    editSettingsAction.setShortcut(QtGui.QKeySequence(settings.shortcut_editSettings))
    editSettingsAction.setStatusTip(self.tr('Edit settings'))
    editSettingsAction.triggered.connect(self.editSettings)
    self.addAction(editSettingsAction)

    saveSettingsAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'document-save', QtGui.QIcon(f'{_icons_path}/document-save.svg')
      ),
      self.tr('&Save settings'),
      self,
    )
    saveSettingsAction.setShortcut(QtGui.QKeySequence(settings.shortcut_saveSettings))
    saveSettingsAction.setStatusTip(self.tr('Save current settings'))
    saveSettingsAction.triggered.connect(settings.write)
    self.addAction(saveSettingsAction)

    self.autoSaveSettingsAction = QtGui.QAction(
      QtGui.QIcon.fromTheme(
        'document-save', QtGui.QIcon(f'{_icons_path}/document-save.svg')
      ),
      self.tr('&Auto save settings'),
      self,
      checkable=True,
    )
    self.autoSaveSettingsAction.setShortcut(
      QtGui.QKeySequence(settings.shortcut_autoSaveSettings)
    )
    self.autoSaveSettingsAction.setStatusTip(
      self.tr('Automatically save settings on exit')
    )
    self.autoSaveSettingsAction.setChecked(settings.auto_save_settings)
    self.autoSaveSettingsAction.triggered.connect(self.toggleAutoSaveSettings)
    self.addAction(self.autoSaveSettingsAction)

    # Menubar
    self.menu = self.menuBar()
    # Clearing menu before repopulating it (in case of UI re-init)
    self.menu.clear()
    self.menu.setVisible(settings.menuVisible)
    fileMenu = self.menu.addMenu(self.tr('&File'))
    fileMenu.addAction(newDocAction)
    fileMenu.addAction(openAction)
    fileMenu.addAction(reloadAction)
    fileMenu.addAction(previewAction)
    fileMenu.addAction(saveAction)
    fileMenu.addAction(saveAsAction)
    fileMenu.addAction(exitAction)

    editMenu = self.menu.addMenu(self.tr('&Edit'))
    editMenu.addAction(undoAction)
    editMenu.addAction(redoAction)
    editMenu.addAction(selectAllAction)
    editMenu.addAction(unselectAllAction)
    editMenu.addAction(copyAction)
    editMenu.addAction(cutAction)
    editMenu.addAction(pasteAction)
    editMenu.addAction(deleteAction)
    editMenu.addAction(editDocAction)
    editMenu.addAction(editAction)
    editMenu.addAction(groupAction)
    editMenu.addAction(ungroupAction)

    insertMenu = self.menu.addMenu(self.tr('&Insert'))
    insertMenu.addAction(newTextAction)
    insertMenu.addAction(newImageAction)
    insertMenu.addAction(newTikzPictureAction)
    insertMenu.addAction(newArrowAction)
    insertMenu.addAction(newFrameAction)
    for tempAct in newFrameTemplateActions:
      insertMenu.addAction(tempAct)
    insertMenu.addAction(saveTemplateAction)
    insertMenu.addAction(deleteTemplateAction)

    moveMenu = self.menu.addMenu(self.tr('&Depth'))
    moveMenu.addAction(moveUpAction)
    moveMenu.addAction(moveDownAction)
    moveMenu.addAction(moveTopAction)
    moveMenu.addAction(moveBottomAction)

    configMenu = self.menu.addMenu(self.tr('&Configuration'))
    configMenu.addAction(self.activateGridAction)
    configMenu.addAction(self.activateObjGuidesAction)
    configMenu.addAction(self.showHiddenObjectsAction)
    configMenu.addAction(self.hideMenuAction)
    configMenu.addAction(editSettingsAction)
    configMenu.addAction(saveSettingsAction)
    configMenu.addAction(self.autoSaveSettingsAction)

    # Page bar (bottom of main window, combobox for the page, frame and framePage)
    # spacer used to center widgets on pagebar
    pagespaceL = QtWidgets.QWidget()
    pagespaceL.setSizePolicy(
      QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
    )
    pagespaceR = QtWidgets.QWidget()
    pagespaceR.setSizePolicy(
      QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
    )
    frameLabel = QtWidgets.QLabel(self)
    frameLabel.setText(self.tr('Frame'))
    pageLabel = QtWidgets.QLabel(self)
    pageLabel.setText(self.tr('Page'))
    framePageLabel = QtWidgets.QLabel(self)
    framePageLabel.setText(self.tr('Page in Frame'))
    self.frameCombo = QtWidgets.QComboBox(self)
    self.frameCombo.currentIndexChanged.connect(self.changeFrame)
    self.pageCombo = QtWidgets.QComboBox(self)
    self.pageCombo.currentIndexChanged.connect(self.changePage)
    self.framePageCombo = QtWidgets.QComboBox(self)
    self.framePageCombo.currentIndexChanged.connect(self.changeFramePage)

    # Remove pagebar before reinserting it (in case of UI re-init)
    if hasattr(self, 'pagebar'):
      self.pagebar.visibilityChanged.disconnect()
      self.removeToolBar(self.pagebar)
    self.pagebar = QtWidgets.QToolBar(self.tr('Pagebar'), self)
    self.addToolBar(QtCore.Qt.BottomToolBarArea, self.pagebar)
    self.pagebar.setVisible(settings.pagebarVisible)
    self.pagebar.visibilityChanged.connect(self.togglePagebar)
    self.pagebar.setFloatable(False)
    self.pagebar.setMovable(False)
    self.pagebar.setStyleSheet('QToolBar{spacing:5px;}')
    self.pagebar.addWidget(pagespaceL)
    self.pagebar.addAction(previousPageAction)
    self.pagebar.addWidget(frameLabel)
    self.pagebar.addWidget(self.frameCombo)
    self.pagebar.addSeparator()
    self.pagebar.addWidget(pageLabel)
    self.pagebar.addWidget(self.pageCombo)
    self.pagebar.addSeparator()
    self.pagebar.addWidget(framePageLabel)
    self.pagebar.addWidget(self.framePageCombo)
    self.pagebar.addAction(nextPageAction)
    self.pagebar.addWidget(pagespaceR)

  def resetUI(self):
    self.initMenus()
    self.makeFrameList()
    self.makePageList()
    self.makeFramePageList()
    self.setIndexes()
    self.painterArea.repaint()

  def resizeEvent(self, event):
    """Things to do when the user resize the window"""
    # Most important resizing consequences are treated by the painterArea resizeEvent
    # Here we just save the new size in the application settings
    settings.window_width = self.width()
    settings.window_height = self.height()

  def newDoc(self):
    """Start a new document from scratch"""
    # If a doc is already opened
    # check if all changes have been saved
    # before starting another one
    if self.document and not self.checkSaved():
      return False
    # Re-init variables
    self.filename = None
    self.folder = None
    self.tmpFile = None
    self.frame = 0
    self.page = 0
    self.framePage = 0
    # Create new doc with a single frame (single page)
    self.document = Document(
      settings.default_header,
      settings.default_footer,
      [Frame(title=settings.default_newdoc_frame_title)],
      settings.default_paper_w,
      settings.default_paper_h,
      1,
    )
    # Re-adjust the size of the painter area
    self.painterArea.setSize()
    # Init of frame, page, framePage combobox
    self.makeFrameList()
    self.makePageList()
    self.makeFramePageList()
    self.setIndexes()
    # Re-Init of undo/redo stack
    self.undoStack = UndoStack(self)
    # Show a preview of the document
    self.previewCurrentFrame()
    return True

  def openDialog(self):
    """Dialog asking for a tex file to open"""
    # Check if all changes have been saved in current doc
    # before opening another one
    if self.checkSaved():
      # Prevent actions in main window (see painter.py) while the dialog is shown
      # Reset mouse cursor to default (remove resize, move... cursors)
      QtWidgets.QApplication.restoreOverrideCursor()
      filename = QtWidgets.QFileDialog.getOpenFileName(
        self, self.tr('Open File'), '.', 'TeX (*.tex)'
      )[0]
      if filename != '':
        # Re-init tmpFile backup
        if self.tmpFile:
          os.remove(self.tmpFile)
          self.tmpFile = None
        # Open the file
        self.openFile(filename)
        # Re-init undo/redo stack
        self.undoStack = UndoStack(self)

  def openFile(self, filename, page=0, reloadAfterSave=False):
    """Open existing tex file given in argument"""
    self.isCompiling = True
    self.painterArea.repaint()
    # Separate folder and filename
    spFolderFile = re.split(r'/', filename)
    folder = r'/'.join(spFolderFile[:-1])
    if folder == '':
      folder = os.getcwd()
    folder += '/'
    # Change working directory to the tex file directory
    os.chdir(folder)
    # Store current folder
    self.folder = folder
    # Separate filename and extension (.tex)
    spNameExt = re.split(r'\.', spFolderFile[-1])
    if spNameExt[-1] == 'tex':
      self.filename = '.'.join(spNameExt[:-1])
    else:
      self.filename = spFolderFile[-1]
    # Compile the tex file
    if not self.compileTeX():
      # If the pdf was not produced (something is broken in the tex file)
      # Ask the user what to do
      self.compileIssueDialog(reloadAfterSave)
      self.isCompiling = False
      self.painterArea.repaint()
      return
    # Re-init document
    self.document = Document()
    # Load document infos (frames, text blocks, images...) from tex and log files
    try:
      self.document.readDoc(self.filename)
    except parsing.MissingBEdLatexPackageError:
      msg = QtWidgets.QMessageBox(self)
      msg.setIcon(QtWidgets.QMessageBox.Warning)
      msg.setWindowTitle(self.tr('LaTeX error detected'))
      msg.setText(
        self.tr(
          'BEd can only load LaTeX files that include the bed LaTeX package. '
          'Please add "\\usepackage{bed}" in the preamble of the file and retry.'
        )
      )
      msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
      msg.exec_()
      self.filename = None
      self.document = None
      self.frame = None
      self.isCompiling = False
      self.painterArea.repaint()
      return
    # Re-adjust the size of the painter area
    self.painterArea.setSize()
    self.page = min(page, self.document.npages - 1)
    # Init of frame, page, framePage combobox
    self.frameFromPage()
    self.makeFrameList()
    self.makePageList()
    self.makeFramePageList()
    self.setIndexes()
    # If tmpFile does not exist create it
    if not self.tmpFile:
      self.tmpFile = tempfile.NamedTemporaryFile(prefix='bed-', suffix='.tex').name
    # Copy current working tex in tmpFile
    os.system('cp "' + self.filename + '.tex" ' + self.tmpFile)
    self.isCompiling = False
    self.painterArea.repaint()

  def compileTeX(self):
    """Compile the current tex file to produce the pdf (system command "pdflatex")
    Check if pdf is produced
    """
    if self.filename:
      # remove old pdf
      if os.path.isfile(self.filename + '.pdf'):
        os.remove(self.filename + '.pdf')
      # compile tex
      os.system(
        settings.latex_cmd
        + ' "'
        + self.filename
        + '.tex" | grep Error > '
        + self.compileOutput
      )
      # Return if the pdf has been produced and no errors have been detected
      return (
        os.path.isfile(self.filename + '.pdf')
        and os.stat(self.compileOutput).st_size == 0
      )
    return False

  def compileIssueDialog(self, reloadAfterSave=False):
    """Ask the user what to do if TeX file compiling did not work"""
    # Reset mouse cursor to default (remove resize, move... cursors)
    QtWidgets.QApplication.restoreOverrideCursor()
    msg = QtWidgets.QMessageBox(self)
    msg.setIcon(QtWidgets.QMessageBox.Warning)
    msg.setWindowTitle(self.tr('LaTeX error detected'))
    with open(self.compileOutput) as cf:
      err = cf.read()
    # Get the previous working tex file if it exists
    if self.tmpFile:
      # Copy backup in working directory
      os.system('cp ' + self.tmpFile + ' "' + self.filename + '.tex.bak"')
      msg.setText(
        self.tr(
          'The document could not be compiled. '
          'The previous saved version of the document has been copied '
          'in the current directory with a .bak extension. '
          'Do you want to reset the document to this backup '
          '(you will lose all changes made) '
          'or continue to edit the current non-working version to try and fix it?'
        )
        + f'\n\n{err}'
      )
      msg.setStandardButtons(QtWidgets.QMessageBox.Reset | QtWidgets.QMessageBox.Retry)
      answer = msg.exec_()
      if answer == QtWidgets.QMessageBox.Reset:
        # If user ask for reset to previous backup
        # erase the tex file with the backup and reload
        os.system('mv "' + self.filename + '.tex.bak" "' + self.filename + '.tex"')
        self.openFile(self.filename, self.page)
        # Re-init undo/redo stack
        self.undoStack = UndoStack(self)
      # else there is nothing to do,
      # the user need to fix the issue to be able to compile again
      # and get a preview of changes
    elif reloadAfterSave:
      # Case 1)
      msg.setText(
        self.tr(
          'The document could not be compiled and no backup of the TeX file was found. '
          'Fix the issue and retry to compile.' + f'\n\n{err}'
        )
      )
      msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
      answer = msg.exec_()
    else:
      # Case 2)
      msg.setText(
        self.tr(
          'The document could not be open in BEd (LaTeX compiling issue). '
          'You will have to fix the issue in an external editor '
          'and retry and open the document in BEd.'
        )
        + f'\n\n{err}'
      )
      msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
      answer = msg.exec_()
      self.filename = None
      self.document = None
      self.frame = None

  def _saveTeXandReload(self):
    """Save the current document in self.filename
    and reload it at page self.page
    """
    # self.filename and self.page must be defined
    self.document.writeTeX(self.filename + '.tex')
    # Re-load document from tex
    if settings.auto_reload:
      self.openFile(self.filename + '.tex', self.page, reloadAfterSave=True)
    # Remember that the saved version is the current one (see undo.py)
    self.undoStack.updateModified()

  def reloadTeX(self):
    """Reload current document from the last saved tex file"""
    if self.filename and self.checkSaved():
      self.openFile(self.filename + '.tex', self.page, reloadAfterSave=True)
      # Remember that the saved version is the current one (see undo.py)
      self.undoStack.updateModified()

  def previewCurrentFrame(self):
    """Compile and load current frame only"""
    if self.document:
      self.isCompiling = True
      self.painterArea.repaint()
      # Create tmp directory
      single_frame_dir_object = tempfile.TemporaryDirectory(prefix='bed-sfp-')
      single_frame_dir = single_frame_dir_object.name
      # Create tmp tex file with a single frame in it
      single_frame_file = single_frame_dir + '/preview.tex'
      self.document.writeTeX(single_frame_file, single_frame=self.frame)
      # Compile it (output to tmp directory)
      for _ in range(2):
        os.system(
          settings.latex_cmd
          + ' -output-directory="'
          + single_frame_dir
          + '"  "'
          + single_frame_file
          + '" | grep Error > '
          + self.compileOutput
        )
      # Check if the pdf has been produced and no errors have been detected
      if (
        os.path.isfile(single_frame_file[:-4] + '.pdf')
        and os.stat(self.compileOutput).st_size == 0
      ):
        # Load the compiled preview
        before = self.document.frames[self.frame].before
        firstpage = self.document.frames[self.frame].firstpage
        pdf = self.document.loadpdf(single_frame_file[:-4] + '.pdf')
        xmlDoc = parsing.parseall(single_frame_file[:-4])
        paper_w = float(xmlDoc.get('paper_w'))
        paper_h = float(xmlDoc.get('paper_h'))
        paper_ratio = paper_w / paper_h
        frame_npages = int(xmlDoc[0].get('n_pages'))
        if self.framePage >= frame_npages:
          dfpages = frame_npages - self.framePage - 1
          self.page += dfpages
          self.framePage += dfpages
        dnpages = frame_npages - self.document.frames[self.frame].npages
        self.document.frames[self.frame].readxml(xmlDoc[0], pdf, paper_ratio, paper_h)
        self.document.frames[self.frame].before = before
        self.document.frames[self.frame].firstpage = firstpage
        for fr in range(self.frame + 1, self.document.nframes):
          self.document.frames[fr].firstpage += dnpages
        self.document.npages += dnpages
        self.makePageList()
        self.makeFramePageList()
        self.setIndexes()
      else:
        QtWidgets.QApplication.restoreOverrideCursor()
        msg = QtWidgets.QMessageBox(self)
        msg.setIcon(QtWidgets.QMessageBox.Warning)
        msg.setWindowTitle(self.tr('LaTeX error detected'))
        with open(self.compileOutput) as cf:
          err = cf.read()
        msg.setText(
          self.tr(
            'The current frame could not be compiled. '
            'Fix the issue and retry to compile.'
          )
          + f'\n\n{err}'
        )
        msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
        msg.exec_()
      # Cleaning tmp directory
      single_frame_dir_object.cleanup()
      self.isCompiling = False
      self.painterArea.repaint()

  def save(self):
    """Save action
    To be called by saveAction.trigger.connect
    """
    if self.document:
      # If the filename is not defined then ask the user (save as...)
      if not self.filename:
        return self.saveAs()
      self._saveTeXandReload()
      # return whether the file has been saved
      return True
    return False

  def saveAs(self):
    """Dialog asking for a tex file to save current document
    To be called by saveAsAction.trigger.connect
    """
    if self.document:
      # Reset mouse cursor to default (remove resize, move... cursors)
      QtWidgets.QApplication.restoreOverrideCursor()
      filename = QtWidgets.QFileDialog.getSaveFileName(
        self, self.tr('Save File'), '.', 'TeX (*.tex)'
      )[0]
      if filename != '':
        # Separate folder and file
        spFolderFile = re.split(r'/', filename)
        folder = r'/'.join(spFolderFile[:-1])
        if folder == '':
          folder = os.getcwd()
        folder += '/'
        # Change working folder to the tex file folder
        os.chdir(folder)
        # Store current working path
        self.folder = folder
        # Separate filename and extension
        spNameExt = re.split(r'\.', spFolderFile[-1])
        if spNameExt[-1] == 'tex':
          self.filename = '.'.join(spNameExt[:-1])
        else:
          self.filename = spFolderFile[-1]
        # Save and reload
        self._saveTeXandReload()
        # Return wether the document has been saved
        return True
    return False

  def checkSaved(self):
    """Check if all changes have been saved in current document"""
    if not self.document or not self.modified:
      return True
    # There are unsaved changes
    # Show a dialog asking the user whether to save them
    # Reset mouse cursor to default (remove resize, move... cursors)
    QtWidgets.QApplication.restoreOverrideCursor()
    msg = QtWidgets.QMessageBox(self)
    msg.setIcon(QtWidgets.QMessageBox.Warning)
    msg.setWindowTitle(self.tr('Unsaved changes'))
    msg.setText(self.tr('Current document has been modified. Save modifications?'))
    msg.setStandardButtons(
      QtWidgets.QMessageBox.Save
      | QtWidgets.QMessageBox.Discard
      | QtWidgets.QMessageBox.Cancel
    )
    answer = msg.exec_()
    # Analyse the answer
    # If Cancel button, then checkSaved return false to warn the calling function
    if answer == QtWidgets.QMessageBox.Cancel:
      return False

    # # If the user ask for saving, launch saveFile
    # # If there is a problem with saving, return false to warn the calling function
    # # Otherwise (the file is saved or the user did not care) return True
    return answer != QtWidgets.QMessageBox.Save or self.save()
    #
    # # If the user ask for saving, launch saveFile
    # if answer == QtWidgets.QMessageBox.Save:
    #   if not self.save():
    #     # If there is a problem with saving, return false to warn the calling function
    #     return False
    # # Otherwise (the file is saved or the user did not care) return True
    # return True

  def closeEvent(self, event):
    """Things to do before closing app"""
    # Check if all changes to current document have been saved
    if self.checkSaved():
      # If its ok then write modified settings to the config file
      if settings.auto_save_settings:
        settings.write()
      # remove temp files
      if self.tmpFile:
        os.remove(self.tmpFile)
      if os.path.isfile(self.compileOutput):
        os.remove(self.compileOutput)
      # and close app
      event.accept()
    else:
      # If there are unsaved changed and the user canceled the closing process
      # then cancel it
      event.ignore()

  def makeFrameList(self):
    """Fill-in the combobox listing frames"""
    if self.document:
      # Remember that we are filling comboboxes
      # And that changes to them should not trigger actions
      self.fillingCombo = True
      self.frameCombo.clear()
      smxf = str(self.document.nframes)
      for f in range(self.document.nframes):
        self.frameCombo.addItem(str(f + 1) + r'/' + smxf)
      self.fillingCombo = False

  def makePageList(self):
    """Fill-in the combobox listing pages"""
    if self.document:
      # Remember that we are filling comboboxes
      # And that changes to them should not trigger actions
      self.fillingCombo = True
      self.pageCombo.clear()
      smxp = str(self.document.npages)
      for p in range(self.document.npages):
        self.pageCombo.addItem(str(p + 1) + r'/' + smxp)
      self.fillingCombo = False

  def makeFramePageList(self):
    """Fill-in the combobox listing framepages"""
    if self.document:
      # Remember that we are filling comboboxes
      # And that changes to them should not trigger actions
      self.fillingCombo = True
      self.framePageCombo.clear()
      smxfp = str(self.document.frames[self.frame].npages)
      for p in range(self.document.frames[self.frame].npages):
        self.framePageCombo.addItem(str(p + 1) + r'/' + smxfp)
      self.fillingCombo = False

  def setIndexes(self):
    """Set indexes in the three frame/page combobox
    to correspond to the current state (frame,page,framepage)
    """
    if self.document:
      # Remember that we are filling comboboxes
      # And that change to them should not trigger actions
      self.fillingCombo = True
      self.frameCombo.setCurrentIndex(self.frame)
      self.pageCombo.setCurrentIndex(self.page)
      self.framePageCombo.setCurrentIndex(self.framePage)
      self.fillingCombo = False

  def frameFromPage(self):
    """Compute self.frame and self.framePage from the value of self.page"""
    if self.document:
      # Simply browse the frame list from 0 until we reach the good one
      self.frame = 0
      while self.frame < self.document.nframes:
        self.framePage = self.page - self.document.frames[self.frame].firstpage
        if 0 <= self.framePage <= self.document.frames[self.frame].npages - 1:
          break
        self.frame += 1

  def previousPage(self):
    """Go to previous page"""
    if self.document and self.page > 0:
      # Check if the previous page is in the same frame or not
      # and adjust frame / framePage accordingly
      self.page -= 1
      if self.framePage > 0:
        self.framePage -= 1
      else:
        self.frame -= 1
        self.framePage = self.document.frames[self.frame].npages - 1
        # If the frame is different
        # The framePage combobox must be re-filled
        self.makeFramePageList()
      # Re-select the good indexes in the 3 frame/page comboboxes
      self.setIndexes()
      self.painterArea.repaint()

  def nextPage(self):
    """Go to next page"""
    # Check if the next page is in the same frame or not
    # and adjust frame / framePage accordingly
    if self.document and self.page < self.document.npages - 1:
      self.page += 1
      if self.framePage < self.document.frames[self.frame].npages - 1:
        self.framePage += 1
      else:
        self.frame += 1
        self.framePage = 0
        # If the frame is different
        # The framePage combobox must be re-filled
        self.makeFramePageList()
      # Re-select the good indexes in the 3 frame/page comboboxes
      self.setIndexes()
      self.painterArea.repaint()

  def changeFrame(self):
    """Frame change (triggered by a user change in the frame combobox)"""
    # Check if this is really the user
    # and not an actions triggered by the filling functions
    # Check if the user really changed something
    if not self.fillingCombo and self.frameCombo.currentIndex() != self.frame:
      # Get frame from selection in combobox
      self.frame = self.frameCombo.currentIndex()
      self.framePage = 0
      # Adjust page number
      self.page = self.document.frames[self.frame].firstpage
      # Re-fill framePage combobox
      self.makeFramePageList()
      # Re-select the good indexes in all comboboxes
      self.setIndexes()
      self.painterArea.repaint()

  def changePage(self):
    """Page change (triggered by a user change in the page combobox)"""
    # Check if this is really the user
    # and not an actions triggered by the filling functions
    # Check if the user really changed something
    if not self.fillingCombo and self.pageCombo.currentIndex() != self.page:
      # Get page from selection in combobox
      self.page = self.pageCombo.currentIndex()
      # Find the corresponding frame and see if it is different from the current one
      f = self.frame
      self.frameFromPage()
      if self.frame != f:
        # Re-fill the framePage combobox if the frame changed
        self.makeFramePageList()
      # Re-select the good indexes in all comboboxes
      self.setIndexes()
      self.painterArea.repaint()

  def changeFramePage(self):
    """FramePage change (triggered by a user change in the framePage combobox)"""
    # Check if this is really the user
    # and not an actions triggered by the filling functions
    # Check if the user really changed something
    if not self.fillingCombo and self.framePageCombo.currentIndex() != self.framePage:
      self.framePage = self.framePageCombo.currentIndex()
      # Compute new page
      self.page = self.document.frames[self.frame].firstpage + self.framePage
      # Re-select the good indexes in all comboboxes
      self.setIndexes()
      self.painterArea.repaint()

  def newFrame(self):
    """Insert a new frame just after the current one"""
    if self.document:
      # Add a temporary snapshot in the undo/redo stack
      # This snapshot can be confirmed (really put in the stack),
      # or canceled (completely forgotten)
      # if the user cancel the creation before the end of the creation process
      self.undoStack.doTemp()
      # Create the frame with default properties (from config file)
      newfr = Frame(
        settings.default_frame_title,
        [],
        [],
        [None],
        self.document.frames[self.frame].firstpage
        + self.document.frames[self.frame].npages,
        1,
      )
      # Adjust the page numbering of subsequent frames
      for f in range(self.frame + 1, self.document.nframes):
        self.document.frames[f].firstpage += 1
      # Adjust document total number of frames and pages
      self.document.nframes += 1
      self.document.npages += 1
      # Adjust currently selected frame to display the newly created one
      self.frame += 1
      # Insert the new frame at the right place in the document frame list
      self.document.frames.insert(self.frame, newfr)
      # Recompute the page number from frame (framePage is 0)
      self.framePage = 0
      self.page = self.document.frames[self.frame].firstpage
      # Re-fill frame/page comboboxes
      self.makeFrameList()
      self.makePageList()
      self.makeFramePageList()
      # Re-select the good indexes for all combobox
      self.setIndexes()
      self.painterArea.repaint()
      # Edit the newly created frame properties (edit dialog)
      self.editAny(new=True)

  def newFrameTemplate(self, templateID):
    """Insert a new frame just after the current one"""
    if self.document:
      # Add a snapshot in the undo/redo stack
      self.undoStack.doTemp()
      try:
        # Create the frame from the template
        spbf = settings.templates[2 * templateID + 1].split('\\begin{frame}')
        before = spbf[0]
        title = spbf[1].split('\\end{frame}')[0]
        newfr = Frame(
          title,
          [],
          [],
          [None],
          self.document.frames[self.frame].firstpage
          + self.document.frames[self.frame].npages,
          1,
          before,
        )
        # Adjust the page numbering of subsequent frames
        for f in range(self.frame + 1, self.document.nframes):
          self.document.frames[f].firstpage += 1
        # Adjust document total number of frames and pages
        self.document.nframes += 1
        self.document.npages += 1
        # Adjust currently selected frame to display the newly created one
        self.frame += 1
        # Insert the new frame at the right place in the document frame list
        self.document.frames.insert(self.frame, newfr)
        # Recompute the page number from frame (framePage is 0)
        self.framePage = 0
        self.page = self.document.frames[self.frame].firstpage
        # Re-fill frame/page comboboxes
        self.makeFrameList()
        self.makePageList()
        self.makeFramePageList()
        # Re-select the good indexes for all combobox
        self.setIndexes()
        self.previewCurrentFrame()
        self.undoStack.confirmTemp()
      except Exception as e:
        print(e)
        self.undoStack.reloadTemp()

  def saveTemplate(self):
    """Save the current frame as a template"""
    if self.document and self.document.nframes > 0:
      # Ask for a name
      QtWidgets.QApplication.restoreOverrideCursor()
      dialog = QtWidgets.QInputDialog(self, QtCore.Qt.Tool)
      dialog.setWindowTitle(self.tr('New template'))
      dialog.setLabelText(self.tr('Template name'))
      dialog.setTextValue(self.tr('Template') + f' {len(settings.templates) // 2 + 1}')
      dialog.resize(300, dialog.height())
      ok = dialog.exec_()
      if ok:
        settings.templates.extend(
          [dialog.textValue(), striplines(self.document.frames[self.frame].writeTeX())]
        )
        # Re-init UI to update templates menu
        self.resetUI()

  def deleteTemplate(self):
    """Delete one of the saved templates"""
    if settings.templates != []:
      # Ask which templates should be deleted
      QtWidgets.QApplication.restoreOverrideCursor()
      dialog = QtWidgets.QInputDialog(self, QtCore.Qt.Tool)
      dialog.setWindowTitle(self.tr('Delete template'))
      dialog.setLabelText(self.tr('Template name'))
      dialog.setComboBoxItems(settings.templates[::2])
      dialog.resize(300, dialog.height())
      ok = dialog.exec_()
      if ok:
        index = settings.templates.index(dialog.textValue())
        settings.templates.pop(index)
        settings.templates.pop(index)
        self.resetUI()

  def newText(self):
    """Insert a new text in the current frame"""
    if self.document:
      # Add a temporary snapshot in the undo/redo stack
      # This snapshot can be confirmed (really put in the stack),
      # or canceled (completely forgotten)
      # if the user cancel the creation before the end of the creation process
      self.undoStack.doTemp()
      # Unselect any currently selected object in the frame
      for gel in (
        self.document.frames[self.frame].elements
        + self.document.frames[self.frame].groups
      ):
        gel.selected = False
      # By default the new text will appear in all pages of the frame (framePages)
      pages = []
      for p in range(self.document.frames[self.frame].npages):
        pages.append(p)
      # Create the new text with default properties (from config file)
      text = Text(
        settings.default_text_x,
        settings.default_text_y,
        settings.default_text_w,
        settings.default_text_h,
        minh=settings.default_text_h,
        text=settings.default_text_text,
        pagesList=pages,
        selected=True,
        paper_ratio=self.document.paper_ratio,
      )
      # Add it at the end of the frame element list
      self.document.frames[self.frame].elements.append(text)
      self.painterArea.repaint()
      # Edit the newly created text properties (edit dialog)
      self.editAny(new=True)

  def newImage(self):
    """Insert a new image in the current frame"""
    if self.document:
      # Add a temporary snapshot in the undo/redo stack
      # This snapshot can be confirmed (really put in the stack),
      # or canceled (completely forgotten)
      # if the user cancel the creation before the end of the creation process
      self.undoStack.doTemp()
      # Unselect any currently selected object in the frame
      for gel in (
        self.document.frames[self.frame].elements
        + self.document.frames[self.frame].groups
      ):
        gel.selected = False
      # By default the new image will appear in all pages of the frame (framePages)
      pages = []
      for p in range(self.document.frames[self.frame].npages):
        pages.append(p)
      # Create the new image with default properties (from config file)
      image = Image(
        settings.default_image_x,
        settings.default_image_y,
        settings.default_image_w,
        settings.default_image_h,
        pagesList=pages,
        selected=True,
        paper_ratio=self.document.paper_ratio,
      )
      # Add it at the end of the frame element list
      self.document.frames[self.frame].elements.append(image)
      self.painterArea.repaint()
      # Edit the newly created image properties (edit dialog)
      self.editAny(new=True)

  def newTikzPicture(self):
    """Insert a new Tikz picture in the current frame"""
    if self.document:
      # Add a temporary snapshot in the undo/redo stack
      # This snapshot can be confirmed (really put in the stack),
      # or canceled (completely forgotten)
      # if the user cancel the creation before the end of the creation process
      self.undoStack.doTemp()
      # Unselect any currently selected object in the frame
      for gel in (
        self.document.frames[self.frame].elements
        + self.document.frames[self.frame].groups
      ):
        gel.selected = False
      # By default the new image will appear in all pages of the frame (framePages)
      pages = []
      for p in range(self.document.frames[self.frame].npages):
        pages.append(p)
      # Create the new image with default properties (from config file)
      tikzpic = TikzPicture(
        settings.default_tikzpicture_x,
        settings.default_tikzpicture_y,
        settings.default_tikzpicture_w,
        settings.default_tikzpicture_h,
        pagesList=pages,
        selected=True,
        paper_ratio=self.document.paper_ratio,
        tikzcmd=settings.default_tikzpicture_cmd,
      )
      # Add it at the end of the frame element list
      self.document.frames[self.frame].elements.append(tikzpic)
      self.painterArea.repaint()
      # Edit the newly created Tikz picture properties (edit dialog)
      self.editAny(new=True)

  def newArrow(self):
    """Insert a new arrow in the current frame"""
    if self.document:
      # Add a temporary snapshot in the undo/redo stack
      # This snapshot can be confirmed (really put in the stack),
      # or canceled (completely forgotten)
      # if the user cancel the creation before the end of the creation process
      self.undoStack.doTemp()
      # Unselect any currently selected object in the frame
      for gel in (
        self.document.frames[self.frame].elements
        + self.document.frames[self.frame].groups
      ):
        gel.selected = False
      # By default the new arrow will appear in all pages of the frame (framePages)
      pages = []
      for p in range(self.document.frames[self.frame].npages):
        pages.append(p)
      # Create the new arrow with default properties (from config file)
      h = settings.default_arrow_lw / self.document.paper_h
      arrow = Arrow(
        settings.default_arrow_x,
        settings.default_arrow_y - h / 2,
        settings.default_arrow_w,
        h,
        pagesList=pages,
        selected=True,
        paper_ratio=self.document.paper_ratio,
        paper_h=self.document.paper_h,
        opt=settings.default_arrow_opt,
      )
      # Add it at the end of the frame element list
      self.document.frames[self.frame].elements.append(arrow)
      self.painterArea.repaint()
      # Edit the newly created arrow properties (edit dialog)
      self.editAny(new=True)

  def group(self):
    """Group selected elements"""
    if self.document:
      # Add a snapshot in the undo/redo stack
      self.undoStack.do()
      # Call the group function of the current frame
      self.document.frames[self.frame].group(self.document.paper_ratio)
    self.painterArea.repaint()

  def ungroup(self):
    """Ungroup selected elements"""
    if self.document:
      # Add a snapshot in the undo/redo stack
      self.undoStack.do()
      # Call the ungroup function of the current frame
      self.document.frames[self.frame].ungroup()
    self.painterArea.repaint()

  def deleteAny(self):
    """Delete selected elements in the current frame
    If nothing is selected, delete the frame
    """
    if self.document:
      # Add a snapshot in the undo/redo stack
      self.undoStack.do()
      # Browse the frame elements/groups
      # To remove selected ones
      # found is True if something is selected
      found = False
      # Reverse the list to be able to remove selected elements
      # without perturbing the for loop
      for el in reversed(self.document.frames[self.frame].elements):
        if el.selected:
          found = True
          self.document.frames[self.frame].elements.remove(el)
      # Reverse the list to be able to remove selected groups
      # without perturbing the for loop
      for g in reversed(self.document.frames[self.frame].groups):
        if g.selected:
          found = True
          for el in g.elements:
            self.document.frames[self.frame].elements.remove(el)
          self.document.frames[self.frame].groups.remove(g)
      # If nothing was selected then the whole frame must be removed from the document
      # Except if it is the only frame (avoid problems)
      if not found and self.document.nframes > 1:
        # Re-adjust the page numberings of subsequent frames
        for f in range(self.frame + 1, self.document.nframes):
          self.document.frames[f].firstpage -= self.document.frames[self.frame].npages
        # Re-adjust the document page and frame numbers
        self.document.npages -= self.document.frames[self.frame].npages
        self.document.nframes -= 1
        # Remove the frame
        del self.document.frames[self.frame]
        # Go to the next frame (first framePage)
        # Except if it was the last frame
        # (in this case go to the previous one, last framePage)
        if self.frame == self.document.nframes:
          self.frame -= 1
          self.framePage = self.document.frames[self.frame].npages - 1
          self.page = self.document.frames[self.frame].firstpage + self.framePage
        else:
          self.framePage = 0
          self.page = self.document.frames[self.frame].firstpage
        # Re-fill comboboxes
        self.makeFrameList()
        self.makePageList()
        self.makeFramePageList()
        # Re-select good indexes in comboboxes
        self.setIndexes()
      self.painterArea.repaint()

  def copyAny(self):
    """Just a version of copyCutAny without args
    To be called by copyAction.trigger.connect
    """
    self.copyCutAny()

  def cutAny(self):
    """Just a version of copyCutAny with a True argument
    To be called by cutAction.trigger.connect
    """
    self.copyCutAny(True)

  def copyCutAny(self, cut=False):
    """Copy or cut selected elements/groups
    If nothing is selected copy or cut the whole frame
    """
    # Init the list of copied elements/groups
    self.copyList = []
    if self.document:
      # Find selected elements and put copies of them in the copyList
      for gel in (
        self.document.frames[self.frame].elements
        + self.document.frames[self.frame].groups
      ):
        if gel.selected:
          self.copyList.append(gel.copy())
      # If nothing was selected, copy the whole frame (and put it in the copyList)
      if self.copyList == []:
        self.copyList = self.document.frames[self.frame].copy()
      # If cut is asked, them remove the selected elements
      if cut:
        self.deleteAny()

  def pasteAny(self):
    """Paste elements/groups/frames contained in the copyList"""
    if self.document:
      # Add a snapshot in the undo/redo stack
      self.undoStack.do()
      # Check if the copyList is a list of elements/groups or a whole frame
      if type(self.copyList) is Frame:
        # insert the frame just after the current one
        self.copyList.firstpage = (
          self.document.frames[self.frame].firstpage
          + self.document.frames[self.frame].npages
        )
        for f in range(self.frame + 1, self.document.nframes):
          self.document.frames[f].firstpage += self.copyList.npages
        self.frame += 1
        self.document.nframes += 1
        self.document.npages += self.copyList.npages
        self.document.frames.insert(self.frame, self.copyList)
        self.page = self.document.frames[self.frame].firstpage
        self.framePage = 0
        self.makeFrameList()
        self.makePageList()
        self.makeFramePageList()
        self.setIndexes()
      else:
        # Unselect any selected object in the current slide
        # before copying elements/groups
        for obj in (
          self.document.frames[self.frame].elements
          + self.document.frames[self.frame].groups
        ):
          obj.selected = False
        # Copy each element of the copyList
        for obj in self.copyList:
          # Create a copy of the object (in order to be able to copy it several times)
          cobj = obj.copy()
          if type(obj) is Group:
            # If the copied object is a group
            # add the object to the frame groups list
            # AND add all the contained elements to the current frame elements list
            self.document.frames[self.frame].groups.append(cobj)
            for el in cobj.elements:
              self.document.frames[self.frame].elements.append(el)
          else:
            # If it is a single element
            # just add it to the frame elements list
            self.document.frames[self.frame].elements.append(cobj)
      self.painterArea.repaint()

  def editAnyConnect(self):
    """Just a version of editAny without args
    To be called by editAction.trigger.connect
    """
    self.editAny()

  def editAny(self, new=False):
    """Edit the selected elements/groups one by one
    If nothing is selected, edit the current frame (frame title)
    """
    if self.document:
      if not new:
        # If the element has not just been created
        # add a temporary snapshot in the undo/redo stack
        # This snapshot can be confirmed (really put in the stack),
        # or canceled (completely forgotten)
        # if the user does not really edit anything
        self.undoStack.doTemp()
      found = False
      validate = False
      for gel in (
        self.document.frames[self.frame].elements
        + self.document.frames[self.frame].groups
      ):
        if gel.selected:
          # if the element/group is selected
          # launch the edit dialog for it
          validate = validate or gel.edit(self)
          found = True
      # If nothing was selected
      # launch the frame edit dialog
      if not found:
        validate = self.document.frames[self.frame].edit(self)
      if validate:
        self.undoStack.confirmTemp()
        self.painterArea.repaint()
      else:
        self.undoStack.reloadTemp()

  def editDoc(self):
    """Edit the document properties (header/footer of tex file)"""
    if self.document:
      # Add a temporary snapshot in the undo/redo stack
      # This snapshot can be confirmed (really put in the stack),
      # or canceled (completely forgotten)
      # if the user does not really edit anything
      self.undoStack.doTemp()
      if self.document.edit(self):
        self.undoStack.confirmTemp()
        self.painterArea.repaint()
      else:
        self.undoStack.reloadTemp()

  def moveUp(self):
    """Move the selected elements/groups one notch up (toward foreground)"""
    # Of course, if a selected element is already at the top, we do nothing
    # One difficulty here is that if the two top elements are selected
    # we must not put the second one on top of the first one
    # indE keep track of this kind of pb
    # If a group is selected, all the contained elements will also be moved up
    # Note that foreground correspond to the end of the lists (last drawn)
    if self.document:
      # Add a snapshot in the undo/redo stack
      self.undoStack.do()
      # Look for selected elements (or elements of selected groups)
      indE = 0
      # Reverse lists to treat top elements first
      for cind, el in enumerate(reversed(self.document.frames[self.frame].elements)):
        # compute the index in list from the index in reversed list
        ind = len(self.document.frames[self.frame].elements) - 1 - cind
        if el.selected or (el.group and el.group.selected):
          # Check if it is possible to put the element up
          if cind > indE:
            # do it by swapping the element with the next one in list
            (
              self.document.frames[self.frame].elements[ind],
              self.document.frames[self.frame].elements[ind + 1],
            ) = (
              self.document.frames[self.frame].elements[ind + 1],
              self.document.frames[self.frame].elements[ind],
            )
          indE += 1
      # If nothing was selected move the current frame up (swap with previous one)
      if indE == 0 and self.frame > 0:
        self.document.frames[self.frame].firstpage -= self.document.frames[
          self.frame - 1
        ].npages
        self.document.frames[self.frame - 1].firstpage += self.document.frames[
          self.frame
        ].npages
        self.page -= self.document.frames[self.frame - 1].npages
        self.document.frames[self.frame], self.document.frames[self.frame - 1] = (
          self.document.frames[self.frame - 1],
          self.document.frames[self.frame],
        )
        self.frame -= 1
        self.setIndexes()
      self.painterArea.repaint()

  def moveDown(self):
    """Move the selected elements/groups one notch down (toward background)"""
    # See initial comment in moveUp
    if self.document:
      # Add a snapshot in the undo/redo stack
      self.undoStack.do()
      # Look for selected elements (or elements of selected groups)
      indE = 0
      for ind, el in enumerate(self.document.frames[self.frame].elements):
        if el.selected or (el.group and el.group.selected):
          # Check if it is possible to put the element down
          if ind > indE:
            # do it by swapping the element with the previous one in list
            (
              self.document.frames[self.frame].elements[ind],
              self.document.frames[self.frame].elements[ind - 1],
            ) = (
              self.document.frames[self.frame].elements[ind - 1],
              self.document.frames[self.frame].elements[ind],
            )
          indE += 1
      # If nothing was selected move the current frame down (swap with next one)
      if indE == 0 and self.frame < self.document.nframes - 1:
        self.document.frames[self.frame].firstpage += self.document.frames[
          self.frame + 1
        ].npages
        self.document.frames[self.frame + 1].firstpage -= self.document.frames[
          self.frame
        ].npages
        self.page += self.document.frames[self.frame + 1].npages
        self.document.frames[self.frame], self.document.frames[self.frame + 1] = (
          self.document.frames[self.frame + 1],
          self.document.frames[self.frame],
        )
        self.frame += 1
        self.setIndexes()
      self.painterArea.repaint()

  def moveTop(self):
    """Move the selected elements/groups to top (toward foreground)"""
    # Same as moveUp but with while loops instead of if statements
    if self.document:
      self.undoStack.do()
      indE = 0
      for tcind, el in enumerate(reversed(self.document.frames[self.frame].elements)):
        # make a copy of the index since we will affect it in the while loop
        cind = tcind
        ind = len(self.document.frames[self.frame].elements) - 1 - cind
        if el.selected or (el.group and el.group.selected):
          while cind > indE:
            (
              self.document.frames[self.frame].elements[ind],
              self.document.frames[self.frame].elements[ind + 1],
            ) = (
              self.document.frames[self.frame].elements[ind + 1],
              self.document.frames[self.frame].elements[ind],
            )
            ind += 1
            cind -= 1
          indE += 1
      if indE == 0:
        while self.frame > 0:
          self.document.frames[self.frame].firstpage -= self.document.frames[
            self.frame - 1
          ].npages
          self.document.frames[self.frame - 1].firstpage += self.document.frames[
            self.frame
          ].npages
          self.page -= self.document.frames[self.frame - 1].npages
          self.document.frames[self.frame], self.document.frames[self.frame - 1] = (
            self.document.frames[self.frame - 1],
            self.document.frames[self.frame],
          )
          self.frame -= 1
        self.setIndexes()
      self.painterArea.repaint()

  def moveBottom(self):
    """Move the selected elements/groups one notch down (toward background)"""
    # Same as moveDown but with while loops instead of if statements
    if self.document:
      self.undoStack.do()
      indE = 0
      for tind, el in enumerate(self.document.frames[self.frame].elements):
        # make a copy of the index since we will affect it in the while loop
        ind = tind
        if el.selected or (el.group and el.group.selected):
          while ind > indE:
            (
              self.document.frames[self.frame].elements[ind],
              self.document.frames[self.frame].elements[ind - 1],
            ) = (
              self.document.frames[self.frame].elements[ind - 1],
              self.document.frames[self.frame].elements[ind],
            )
            ind -= 1
          indE += 1

      if indE == 0:
        while self.frame < self.document.nframes - 1:
          self.document.frames[self.frame].firstpage += self.document.frames[
            self.frame + 1
          ].npages
          self.document.frames[self.frame + 1].firstpage -= self.document.frames[
            self.frame
          ].npages
          self.page += self.document.frames[self.frame + 1].npages
          self.document.frames[self.frame], self.document.frames[self.frame + 1] = (
            self.document.frames[self.frame + 1],
            self.document.frames[self.frame],
          )
          self.frame += 1
        self.setIndexes()
      self.painterArea.repaint()

  def undo(self):
    """Undo function"""
    # This is only an intermediate function
    # the real one is implemented in undo.py (undoStack class)
    # This is used to be able to load the ui before defining the undoStack
    if self.document:
      self.undoStack.undo()

  def redo(self):
    """Redo function"""
    # This is only an intermediate function
    # the real one is implemented in undo.py (undoStack class)
    # This is used to be able to load the ui before defining the undoStack
    if self.document:
      self.undoStack.redo()

  def toggleGrid(self):
    """Show/hide the grid and activate/deactivate grid magnetism"""
    # Get the state from the activateGridAction check state
    settings.activate_grid = self.activateGridAction.isChecked()
    self.painterArea.repaint()

  def toggleObjGuides(self):
    """Show/hide the objects guides and activate/deactivate objects magnetism"""
    # Get the state from the activateObjGuidesAction check state
    settings.activate_object_guides = self.activateObjGuidesAction.isChecked()
    self.painterArea.repaint()

  def toggleHiddenObjects(self):
    """Show/hide the hidden objects in the page"""
    # Get the state from the showHiddenObjectsAction check state
    settings.show_hidden_objects = self.showHiddenObjectsAction.isChecked()
    self.painterArea.repaint()

  def toggleMenu(self):
    """Show/hide menubar"""
    # Get the state from the hideMenuAction check state
    settings.menuVisible = self.hideMenuAction.isChecked()
    # Set visibility accordingly
    self.menu.setVisible(settings.menuVisible)

  def togglePagebar(self, visible):
    """Show/hide Pagebar"""
    # Save visibility in settings
    settings.pagebarVisible = visible

  def editSettings(self):
    if settings.edit(self):
      # Re-init UI in case some properties changed
      self.resetUI()

  def toggleAutoSaveSettings(self):
    """Switch on/off auto-save settings on application closing"""
    # Get the state from the autoSaveSettingsAction check state
    settings.auto_save_settings = self.autoSaveSettingsAction.isChecked()

  def errorHandler(self, *args):
    """Show the error and useful info in a message box when an exception is raised"""
    traceback.print_last()
    QtWidgets.QApplication.restoreOverrideCursor()
    msg = QtWidgets.QMessageBox(self)
    msg.setIcon(QtWidgets.QMessageBox.Critical)
    msg.setWindowTitle(self.tr('BEd error'))
    text = self.tr('The following error was detected:')
    trb = '\n'.join(traceback.format_exception(*args))
    text += '<pre>' + html.escape(f'\n{trb}\n') + '</pre>'
    text += self.tr('Try to save your document and restart BEd.')
    text += '<br><br>' + self.tr('Please consider filling a bug report at:')
    text += '<br><a href="https://framagit.org/delisle/bed/-/issues">https://framagit.org/delisle/bed/-/issues</a>'
    msg.setText(text)
    msg.setTextFormat(QtCore.Qt.RichText)
    msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
    msg.exec_()


def main():
  # Make sure latex will find bed.sty
  if os.system('kpsewhich bed.sty'):  # detects error codes (returns non-zero value)
    os.environ['TEXINPUTS'] = f'{os.environ.get("TEXINPUTS","")}:{_latex_path}'
  # Initialise app
  app = QtWidgets.QApplication(sys.argv)
  app.setApplicationName('BEd')
  app.setApplicationDisplayName('BEd')
  app.setDesktopFileName('bed')
  # Get the system language to load translator files
  lang = QtCore.QLocale().system().name().split('_')[0]
  translator = QtCore.QTranslator()
  translator.load(f'{lang}.qm', _tr_path)
  app.installTranslator(translator)
  bed = BEd()
  sys.excepthook = bed.errorHandler

  # Check if optionnal arguments are passed (filename, pagenumber)
  def initdoc():
    if len(sys.argv) > 2:
      bed.initDocument(sys.argv[1], int(sys.argv[2]) - 1)
    elif len(sys.argv) > 1:
      bed.initDocument(sys.argv[1])
    else:
      bed.initDocument()

  # Launch initdoc after application (wait 100 ms)
  QtCore.QTimer.singleShot(100, initdoc)
  sys.exit(app.exec_())



"""
Copyright 2015-2024 Jean-Baptiste Delisle

This file is part of BEd.

BEd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BEd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with BEd.  If not, see <http://www.gnu.org/licenses/>.
"""

# This file defines the UndoStack class, that allow to perform undo/redo actions

from .settings import settings


class State:
  """State class defines a complete document state to be saved in the undo/redo stack"""

  def __init__(self, document=None, frame=0, page=0, framePage=0, modified=False):
    """Initialisation"""
    # Deep copy
    self.document = document.copy()
    self.frame = frame
    self.page = page
    self.framePage = framePage
    self.modified = modified

  def copy(self):
    """Deep copy"""
    return State(self.document, self.frame, self.page, self.framePage, self.modified)


class UndoStack:
  """UndoStack class defines an undo/redo stack to save
  and reload intermediate states of the document
  """

  def __init__(self, parent):
    """Initialisation"""
    self.before_states = []
    self.after_states = []
    self.parent = parent
    self.tempState = None
    self.parent.modified = False

  def getState(self):
    """Get the current state of the document from parent"""
    return State(
      self.parent.document,
      self.parent.frame,
      self.parent.page,
      self.parent.framePage,
      self.parent.modified,
    )

  def setState(self, state):
    """Set the current state of the document to "state" """
    self.parent.document = state.document
    self.parent.frame = state.frame
    self.parent.page = state.page
    self.parent.framePage = state.framePage
    self.parent.modified = state.modified
    self.parent.makeFrameList()
    self.parent.makePageList()
    self.parent.makeFramePageList()
    self.parent.setIndexes()
    self.parent.painterArea.repaint()

  def do(self, current_state=None):
    """Save the current state in the stack before doing a modifications"""
    # Here is the behavior:
    # If the user do actions A, then B then C
    # Then undo action C and B (go back to A state)
    # Then modify the document (action D)
    # The undo stack will be A,B,C,B,A,D
    if not current_state:
      current_state = self.getState()
    self.before_states.append(current_state)
    # If there are some redo actions in after_states
    # We are in the case described at the beginning of the function
    # So we copy states in the after_states list twice (ABCBA..)
    if self.after_states != []:
      for state in reversed(self.after_states[1:]):
        self.before_states.append(state.copy())
      self.before_states += self.after_states
      self.before_states.append(current_state.copy())
      self.after_states = []
    # Remove old preview images from memory (to avoid memory leakage)
    for state in self.before_states[: -settings.Nkeep_previews]:
      state.document.clean_imgs()
    self.parent.modified = True

  def doTemp(self):
    """Do a temporary state save.
    This state can be confirmed later and put in the stack,
    can be reloaded in the document, or completely forgotten and deleted.
    """
    self.tempState = self.getState()

  def confirmTemp(self):
    """Confirm the last temp state"""
    if self.tempState:
      self.do(self.tempState)
      self.tempState = None

  def cancelTemp(self):
    """Cancel (forget) the last temp state"""
    self.tempState = None

  def reloadTemp(self):
    """Reload the last temp state"""
    if self.tempState:
      self.setState(self.tempState)
      self.tempState = None

  def undo(self):
    """Undo function, load the previous state in the undo stack"""
    if self.before_states != []:
      self.after_states.append(self.getState())
      self.setState(self.before_states.pop())

  def redo(self):
    """Redo function, load the next state undo stack"""
    if self.after_states != []:
      self.before_states.append(self.getState())
      self.setState(self.after_states.pop())

  def updateModified(self):
    """Update all states modified field if the document has been saved"""
    self.parent.modified = False
    for state in self.before_states + self.after_states:
      state.modified = True



"""
Copyright 2015-2024 Jean-Baptiste Delisle

This file is part of BEd.

BEd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BEd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with BEd.  If not, see <http://www.gnu.org/licenses/>.
"""

# This file defines the Frame class (Beamer frame)

import xml.etree.ElementTree as ET

from PySide6 import QtCore

from .editor import BedAdjustTextEdit, PropertiesEditor
from .element import Arrow, Image, Text, TikzPicture
from .group import Group
from .helpers import spacelines, striplines
from .settings import settings


class Frame(QtCore.QObject):
  """Frame class represents a complete Beamer frame.
  Contains a title and elements (texts, images, arrows)
  """

  def __init__(self, *args, **kwargs):
    """Initialisation"""
    super().__init__()
    if len(args) > 0 and type(args[0]) is ET.Element:
      self.readxml(*args, **kwargs)
    else:
      self.init(*args, **kwargs)

  def init(
    self,
    title='',
    elements=(),
    groups=(),
    pageimgs=(None,),
    firstpage=0,
    npages=1,
    before='',
  ):
    """Initialisation from parameters.
    This function is used for the copy of a frame,
    so the init of properties is done with a deep copy.
    """
    self.title = title
    # Elements (text/image/arrow)
    self.elements = []
    # Groups of elements (see group.py)
    self.groups = []
    # Deep copy of groups/elements
    dic = {}
    for g in groups:
      ng = g.copy(copy_elements=False)
      self.groups.append(ng)
      dic[g] = ng
    for el in elements:
      self.elements.append(el.copy())
      if el.group:
        ng = dic[el.group]
        ng.addElement(self.elements[-1], compute_geometry=False)
    # First page the frame appears in the document
    self.firstpage = firstpage
    self.npages = npages
    self.pageimgs = list(pageimgs)
    # Latex commands that appear just before the frame
    self.before = before

  def readxml(self, xmlFrame, pageimgs, paper_ratio, paper_h):
    """Initialisation from xml file"""
    # Same as init() but reads the values from the xml tree instead of function args.
    self.title = striplines(xmlFrame.get('title').strip(), 4)
    self.firstpage = int(xmlFrame.get('first_page')) - 1
    self.npages = int(xmlFrame.get('n_pages'))
    self.pageimgs = pageimgs[self.firstpage : self.firstpage + self.npages]
    self.before = striplines(xmlFrame.get('before').strip(), 2)
    self.groups = []
    for _ in range(int(xmlFrame.get('n_groups'))):
      self.groups.append(Group(paper_ratio=paper_ratio))
    self.elements = []
    for xml_element in reversed(xmlFrame):
      if xml_element.tag == 'img':
        self.elements.insert(
          0, Image(xml_element, self.npages, self.pageimgs, paper_ratio)
        )
      elif xml_element.tag == 'tkp':
        self.elements.insert(
          0, TikzPicture(xml_element, self.npages, self.pageimgs, paper_ratio)
        )
      elif xml_element.tag == 'txt':
        self.elements.insert(
          0, Text(xml_element, self.npages, self.pageimgs, paper_ratio)
        )
      elif xml_element.tag == 'arw':
        self.elements.insert(
          0, Arrow(xml_element, self.npages, self.pageimgs, paper_ratio, paper_h)
        )
      if self.elements[0].group < 0:
        self.elements[0].group = None
      else:
        self.groups[self.elements[0].group].addElement(self.elements[0], append=False)

  def copy(self):
    """Deep copy except for pageimgs (pdf preview)"""
    return Frame(
      self.title,
      self.elements,
      self.groups,
      self.pageimgs,
      self.firstpage,
      self.npages,
      self.before,
    )

  def writeTeX(self):
    """Write the TeX extract corresponding to the frame properties"""
    # Try to get correct indenting
    tex = ''
    if self.before.strip():
      tex += spacelines(self.before.strip(), 2) + '\n'
    tex += '  \\begin{frame}'
    first = True
    for line in self.title.split('\n'):
      if first:
        first = False
      else:
        tex += '    '
      tex += line + '\n'
    for el in self.elements:
      tex += el.writeTeX(self.groups) + '\n'
    tex += '  \\end{frame}'
    return tex

  def group(self, paper_ratio):
    """Group selected elements in the frame"""
    g = Group(paper_ratio=paper_ratio)
    tmpg = []
    # Look if groups are selected
    # reverse list to be able to delete groups without perturbing loop
    for gr in reversed(self.groups):
      if gr.selected:
        # Add all elements of the selected group
        tmpg += gr.elements
        # Delete the old group
        self.groups.remove(gr)
    # Look for selected elements
    for el in self.elements:
      if el.selected or el in tmpg:
        # unselect and deactivate hover properties (see refreshHovering)
        el.selected = False
        el.hover = 0
        # Add
        g.addElement(el)
    # If the group is not empty
    if g.elements:
      # Select it
      g.selected = True
      # add it to the frame group list
      self.groups.append(g)

  def ungroup(self):
    """Ungroup selected groups"""
    # Look if groups are selected
    # reverse list to be able to delete groups without perturbing loop
    for g in reversed(self.groups):
      if g.selected:
        # Free all contained elements
        for el in g.elements:
          el.selected = True
          el.group = None
        # delete group from the frame list
        self.groups.remove(g)

  def tmpgroup(self, paper_ratio):
    """Group selected elements/groups in the frame in a temporary group
    Useful for moving multiple elements without creating a permanent group
    see painter.py
    """
    g = Group(paper_ratio=paper_ratio)
    # Look for selected elements/groups
    for gel in self.elements + self.groups:
      if gel.selected:
        # unselect and deactivate hover properties (see refreshHovering)
        gel.selected = False
        gel.hover = 0
        # Add
        g.addElement(gel)
    # If the group is not empty
    if g.elements:
      # Select it
      g.selected = True
      # add it to the frame group list
      self.groups.append(g)

  def untmpgroup(self):
    """Ungroup the temporary group only"""
    # Look if groups are selected
    # reverse list to be able to delete groups without perturbing loop
    # But stop after the first removal
    # (to avoid deleting groups contained in the tmpgroup)
    for g in reversed(self.groups):
      if g.selected:
        # Free all contained elements
        for el in g.elements:
          el.selected = True
          el.group = None
        # delete group from the frame list
        self.groups.remove(g)
        break

  def xGuides(self, framepage):
    """Compute the object guides of the frame along x"""
    # Look for all groups and elements positions to get guides when
    # moving an object in the frame (allow to easily align elements/groups)
    xg = []
    # Only do it if Object guides are activated
    if settings.activate_object_guides:
      # Browse the elements/groups lists
      for gel in self.elements + self.groups:
        # Don't consider selected objects (they will not guide themselves)
        # Also elements in selected groups should not be considered
        if (
          gel.selected
          or (gel.group and gel.group.selected)
          or not (settings.show_hidden_objects or gel.isvisible(framepage))
        ):
          continue
        # Add object guides to the list
        xg += gel.xguides
    # If the grid is activated
    # Add grid guides to the list
    if settings.activate_grid:
      for k in range(settings.Ngrid + 1):
        xg.append(k / settings.Ngrid)
    return xg

  def yGuides(self, framepage):
    """Compute the object guides of the frame along x"""
    # See comments of xGuides
    yg = []
    if settings.activate_object_guides:
      for gel in self.elements + self.groups:
        if (
          gel.selected
          or (gel.group and gel.group.selected)
          or not (settings.show_hidden_objects or gel.isvisible(framepage))
        ):
          continue
        # Add object guides to the list
        yg += gel.yguides
    if settings.activate_grid:
      for k in range(settings.Ngrid + 1):
        yg.append(k / settings.Ngrid)
    return yg

  def draw(self, framepage, painter):
    """Draw the current state of the frame page"""
    # Destination rectangle for the background image
    rect = QtCore.QRect(0, 0, painter.parent.width(), painter.parent.height())
    if self.pageimgs[framepage]:
      painter.drawImage(rect, self.pageimgs[framepage])
    else:
      painter.setBrush(settings.color_background)
      painter.drawRect(rect)
      painter.setBrush(QtCore.Qt.NoBrush)
    # Draw borders
    painter.setPen(settings.pen_grid1)
    painter.drawVline(0)
    painter.drawVline(1)
    painter.drawHline(0)
    painter.drawHline(1)
    # Draw guides
    # Check if there is any selected objects
    if any(gel.selected for gel in self.groups + self.elements):
      # Fixed grid
      if settings.activate_grid:
        for k in range(1, settings.Ngrid):
          if int(4 * k / settings.Ngrid) == 4 * k / settings.Ngrid:
            painter.setPen(settings.pen_grid1)
          elif int(8 * k / settings.Ngrid) == 8 * k / settings.Ngrid:
            painter.setPen(settings.pen_grid2)
          elif int(16 * k / settings.Ngrid) == 16 * k / settings.Ngrid:
            painter.setPen(settings.pen_grid3)
          else:
            painter.setPen(settings.pen_grid4)
          painter.drawVline(k / settings.Ngrid)
          painter.drawHline(k / settings.Ngrid)
      # Object guides
      frame_xguides = self.xGuides(framepage)
      frame_yguides = self.yGuides(framepage)
      if settings.activate_object_guides or settings.activate_grid:
        # Unselected objects guides in background
        painter.setPen(settings.pen_object_guide)
        for gel in self.groups + self.elements:
          if (
            settings.activate_object_guides
            and not gel.group
            and not gel.selected
            and (settings.show_hidden_objects or gel.isvisible(framepage))
          ):
            for xg in gel.xguides:
              painter.drawVline(xg)
            for yg in gel.yguides:
              painter.drawHline(yg)
        # Selected objects guides in foreground
        for gel in self.groups + self.elements:
          if gel.selected and not gel.group:
            gel.refresh_align_guides(frame_xguides, frame_yguides)
            for ng, xg in zip(gel.xguides_names, gel.xguides):
              if ng in gel.current_align_x:
                painter.setPen(settings.pen_aligned_object_guide)
              else:
                painter.setPen(settings.pen_selected_object_guide)
              painter.drawVline(xg)
            for ng, yg in zip(gel.yguides_names, gel.yguides):
              if ng in gel.current_align_y:
                painter.setPen(settings.pen_aligned_object_guide)
              else:
                painter.setPen(settings.pen_selected_object_guide)
              painter.drawHline(yg)
    # Draw elements / groups
    groupscount = [len(grp.elements) for grp in self.groups]
    # Loop over elements
    for el in self.elements:
      if el.isvisible(framepage):
        el.draw(framepage, painter)
        # The groups are drawn at the same time as their uppermost element
        if el.group:
          kgrp = self.groups.index(el.group)
          groupscount[kgrp] -= 1
          if groupscount[kgrp] == 0:
            el.group.draw(framepage, painter)
    # In case we did not draw some group
    # (in particular for the temporary group (tmpgroup))
    for kgrp, grp in enumerate(self.groups):
      if groupscount[kgrp] > 0 and grp.isvisible(framepage):
        grp.draw(framepage, painter)

  def edit(self, parentwindow):
    """Launch the PropertiesEditor (see editor.py) to edit the Frame properties."""
    # Initial values of props (before editing)
    init_values = [self.before, self.title]
    # Define BedWidget for each prop (see editor.py)
    bwbefore = BedAdjustTextEdit(
      init_values[0], label=self.tr('Before frame'), highlight=True
    )
    bwtitle = BedAdjustTextEdit(
      init_values[1], label=self.tr('Title & header'), highlight=True
    )
    bedwidgets = [bwbefore, bwtitle]
    values = PropertiesEditor.getValues(
      parentwindow, bedwidgets, settings, 'frame', self.tr('Frame')
    )

    # Check if values have changed
    if values:
      # Update properties
      self.before = values[0]
      self.title = values[1]
      return True
    return False

  def clean_imgs(self):
    """Clean all preview images from the frame and its elements."""
    self.pageimgs = [None] * self.npages
    for el in self.elements:
      el.clean_imgs()



"""
Copyright 2015-2024 Jean-Baptiste Delisle

This file is part of BEd.

BEd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BEd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with BEd.  If not, see <http://www.gnu.org/licenses/>.
"""

# This file defines TeX parsing functions

import math
import xml.etree.ElementTree as ET


class MissingBEdLatexPackageError(Exception):
  """Raise this when \\usepackage{bed} is missing in the TeX file preamble."""

  def __init__(self):
    super().__init__('Could not find \\usepackage{bed} in TeX file preamble.')


# Generic parsing functions
def _aux_find(string, substring, escape=None):
  """Find the first occurence of a substring (can be a list of possible substrings)
  in a string and return its position (start,end).
  Do not count the occurences preceded by the escape string.
  """
  subs = substring
  if type(subs) is str:
    subs = [subs]
  lsubs = [(len(sub), sub) for sub in subs]
  if escape:
    lesc = len(escape)
  k = 0
  while k < len(string):
    for lsub, sub in lsubs:
      if string[k : k + lsub] == sub:
        return (k, k + lsub)
    if escape and string[k : k + lesc] == escape:
      k += lesc + 1
    else:
      k += 1
  return (-1, -1)


def _aux_findall(string, substring, escape=None):
  """Find all occurences of a substring (can be a list of possible substrings)
  in a string and return their position (list of (start,end)).
  Do not count the occurences preceded by the escape string.
  """
  inds = []
  k = 0
  while k < len(string):
    s, e = _aux_find(string[k:], substring, escape)
    if s < 0:
      break
    inds.append((k + s, k + e))
    k += e
  return inds


def find(string, substring, escape=None, se='start', restrict=None):
  """Find the first occurence of a substring in a string restricted
  to the interval list restrict.
  se="start" -> output only the start position of the substring
  se="end" -> output only the end position of the substring
  otherwise -> output start,end position of the substring
  """
  if not restrict:
    restrict = [[0, len(string)]]
  for a, b in restrict:
    s, e = _aux_find(string[a:b], substring, escape)
    if s >= 0:
      if se == 'start':
        return a + s
      if se == 'end':
        return a + e
      return (a + s, a + e)
  if se in ('start', 'end'):
    return -1
  return (-1, -1)


def findall(string, substring, escape=None, se='start', restrict=None):
  """Find all the occurences of a substring in a string restricted
  to the interval list restrict.
  se="start" -> output only the start position of the substring
  se="end" -> output only the end position of the substring
  otherwise -> output start,end position of the substring
  """
  if not restrict:
    restrict = [[0, len(string)]]
  inds = []
  for a, b in restrict:
    tmp = _aux_findall(string[a:b], substring, escape)
    for s, e in tmp:
      if se == 'start':
        inds.append(a + s)
      elif se == 'end':
        inds.append(a + e)
      else:
        inds.append((a + s, a + e))
  return inds


def match(lstart, lend, start, multilevel=True):
  """Find the ending substring matching the starting substring
  lstart: list of positions of start substrings
  lend: list of positions of end substrings
  start: position of the start substring we want to match
  """
  kstart = 0
  while kstart < len(lstart) and lstart[kstart] < start:
    kstart += 1
  kend = 0
  while kend < len(lend) and lend[kend] <= lstart[kstart]:
    kend += 1
  if multilevel:
    ks = kstart + 1
    ke = kend
    while ke < len(lend) and ke - kend < ks - kstart:
      while ks < len(lstart) and lstart[ks] < lend[ke]:
        ks += 1
      ke += 1
    if ke - kend < ks - kstart:
      return -1
    return lend[ke - 1]
  return lend[kend] if kend < len(lend) else -1


def find_comments(string, start='%', end='\n', escape_start='\\', escape_end=None):
  """Find all the comments in the file.
  Comments are delimited by the start and end substrings
  except when preceded by escape substrings.
  """
  comstart = findall(string, start, escape_start)
  comend = findall(string, end, escape_end, 'end')
  comments = []
  k = 0
  e = 0
  while k < len(comstart):
    if comstart[k] >= e:
      s = comstart[k]
      if comstart[k] == e:
        s = comments[-1][0]
        del comments[-1]
      e = match(comstart[k:], comend, comstart[k], False)
      if e < 0:
        break
      comments.append((s, e))
    k += 1
  return comments


def find_strings(string, delimitor='"', escape='\\', restrict=None):
  """Find all the strings (begin and end delimitor are the same)"""
  ldel = len(delimitor)
  dels = findall(string, delimitor, escape, restrict=restrict)
  if len(dels) % 2 == 1:
    return -1
  strs = []
  for s, e in zip(dels[::2], dels[1::2]):
    strs.append((s, e + ldel))
  return strs


def complementary(intlist, n):
  """Compute the complementary set of interval"""
  oldb = 0
  comp = []
  for a, b in intlist:
    if a > oldb:
      comp.append((oldb, a))
    oldb = b
  if oldb < n:
    comp.append((oldb, n))
  return comp


def intersection(intlist1, intlist2):
  """Compute the intersection between 2 intervals"""
  inter = []
  for a1, b1 in intlist1:
    for a2, b2 in intlist2:
      ia = max(a1, a2)
      ib = min(b1, b2)
      if ia < ib:
        inter.append((ia, ib))
  return inter


################################# Initial parsing of the tex file
def parsetex(jobname, xml_doc):
  """Parse the whole tex file to find frames, elements...
  and construct xml tree of it.
  """
  with open(jobname + '.tex') as f:
    tex = f.read()
  ltex = len(tex)

  # Find commented parts in the file
  coms = find_comments(tex)
  # Take the complementary to find the non-commented parts
  code = complementary(coms, ltex)

  if find(tex, '\\usepackage{bed}', escape='\\', restrict=code, se='start') == -1:
    raise MissingBEdLatexPackageError
  # Find the {document} (begin, end)
  doc_b = find(tex, '\\begin{document}', escape='\\', restrict=code, se='end')
  doc_e = find(tex, '\\end{document}', escape='\\', restrict=code)
  xml_doc.set('header', tex[:doc_b])
  # Find the non-commented parts of the code inside the {document}
  doc_code = intersection(code, [[doc_b, doc_e]])
  # Find the frames (begin, end)
  frames_b = findall(tex, '\\begin{frame}', escape='\\', restrict=doc_code, se='both')
  frames_e = findall(tex, '\\end{frame}', escape='\\', restrict=doc_code, se='both')
  # Explore each Frame
  pfr_e = doc_b
  for (fr_bb, fr_be), (fr_eb, fr_ee) in zip(frames_b, frames_e):
    # Add new frame in the tree
    xml_fr = ET.SubElement(xml_doc, 'frame')
    # Initialisation
    xml_fr.set('n_pages', '1')
    xml_fr.set('n_groups', '0')
    # Code that appear before the frame
    xml_fr.set('before', tex[pfr_e:fr_bb])
    pfr_e = fr_ee
    # Code inside frame
    frame_code = intersection(doc_code, [[fr_be, fr_eb]])
    # Find all the brackets in the frame
    brackets_b = findall(tex, '{', escape='\\', restrict=frame_code)
    brackets_e = findall(tex, '}', escape='\\', restrict=frame_code, se='end')
    # Find all the elements (\txt, \img, \tkp, \arw)
    elems = findall(
      tex,
      ['\\txt', '\\img', '\\tkp', '\\arw'],
      escape='\\',
      restrict=frame_code,
      se='both',
    )
    pel_e = -1
    # Analyse each element
    for el_b, el_e in elems:
      # Create new element
      xml_el = ET.SubElement(xml_fr, tex[el_b + 1 : el_e])
      if pel_e == -1:
        # Find frame title if first element
        xml_fr.set('title', tex[fr_be:el_b])
        xml_el.set('before', '')
      else:
        # Code before element
        xml_el.set('before', tex[pel_e:el_b])
      if tex[el_b:el_e] == '\\arw':
        pel_e = match(brackets_b, brackets_e, el_e)
      else:
        firstb_e = match(brackets_b, brackets_e, el_e)
        pel_e = match(brackets_b, brackets_e, firstb_e)
        if tex[el_b:el_e] == '\\txt':
          xml_el.set('text', tex[firstb_e + 1 : pel_e - 1])
        elif tex[el_b:el_e] == '\\img':
          xml_el.set('name', tex[firstb_e + 1 : pel_e - 1])
        else:
          xml_el.set('tikzcmd', tex[firstb_e + 1 : pel_e - 1])

    # Find title if no elements
    if not elems:
      xml_fr.set('title', tex[fr_be:fr_eb])
      pel_e = fr_eb
    xml_fr.set('footer', tex[pel_e:fr_eb])
  xml_doc.set('footer', tex[fr_ee:])
  xml_doc.set('n_frames', len(frames_b))


def parselog(jobname, xml_doc):
  """Parsing of the log file (after Tex compilation)
  The bed latex package (see ../latex/bed.sty) output some informations to the log file
  """
  # Read all lines in the log
  with open(jobname + '.log', encoding='ISO-8859-1') as f:
    log = f.readlines()
  k = 0
  frame = -1
  framefirstpage = 0
  nelem = 0
  while k < len(log):
    # Check if this is an interesting line (must start with BEd package)
    if log[k][:11] == 'BEd package':
      # Initial informations (paper size)
      if log[k][14:17] == 'beg':
        xml_doc.set('paper_w', log[k + 1].strip()[:-2])
        xml_doc.set('paper_h', log[k + 2].strip()[:-2])
        paperw = float(log[k + 1].strip()[:-2])
        paperh = float(log[k + 2].strip()[:-2])
        paperratio = paperw / paperh
        k += 2
      # Final informations (total number of pages)
      elif log[k][14:17] == 'end':
        xml_doc.set('n_pages', int(log[k + 1]) - 1)
        if frame > -1:
          xml_doc[frame].set('n_pages', int(log[k + 1]) - framefirstpage)
          xml_doc[frame].set('n_elements', nelem)
        k += 1
      elif log[k][14:17] == 'fra':
        # Get frame / firstpage where it appears
        if frame > -1:
          xml_doc[frame].set('n_pages', int(log[k + 2]) - framefirstpage)
          xml_doc[frame].set('n_elements', nelem)
        frame = int(log[k + 1])
        framefirstpage = int(log[k + 2])
        nelem = 0
        xml_doc[frame].set('first_page', framefirstpage)
        xml_doc[frame].set('n_pages', '1')
        k += 2
      else:
        # Element information
        page = int(log[k + 2])
        if page == framefirstpage:
          # Check type of element
          if log[k][14:17] == 'txt':
            # Text
            xml_doc[frame][nelem].set('x', log[k + 3].strip())
            xml_doc[frame][nelem].set('y', log[k + 4].strip())
            xml_doc[frame][nelem].set('w', log[k + 5].strip())
            xml_doc[frame][nelem].set('h', log[k + 6].strip())
            xml_doc[frame][nelem].set('minh', log[k + 7].strip())
            xml_doc[frame][nelem].set('angle', log[k + 8].strip())
            xml_doc[frame][nelem].set('pages', log[k + 9].strip())
            xml_doc[frame][nelem].set('group', log[k + 10].strip())
            xml_doc[frame][nelem].set('align', log[k + 11].strip())
            k += 11
          elif log[k][14:17] == 'img':
            # Image
            xml_doc[frame][nelem].set('x', log[k + 3].strip())
            xml_doc[frame][nelem].set('y', log[k + 4].strip())
            xml_doc[frame][nelem].set('w', log[k + 5].strip())
            xml_doc[frame][nelem].set('h', log[k + 6].strip())
            xml_doc[frame][nelem].set('angle', log[k + 7].strip())
            xml_doc[frame][nelem].set('pages', log[k + 8].strip())
            xml_doc[frame][nelem].set('group', log[k + 9].strip())
            xml_doc[frame][nelem].set('trim', log[k + 10].strip())
            xml_doc[frame][nelem].set('link', log[k + 11].strip())
            xml_doc[frame][nelem].set('origratio', log[k + 12].strip())
            xml_doc[frame][nelem].set(
              'isorigratio', (1 if log[k + 13].strip() == '-1' else 0)
            )
            xml_doc[frame][nelem].set('ukoa', log[k + 14].strip()[1:])
            k += 14
          elif log[k][14:17] == 'tkp':
            # Image
            xml_doc[frame][nelem].set('x', log[k + 3].strip())
            xml_doc[frame][nelem].set('y', log[k + 4].strip())
            xml_doc[frame][nelem].set('w', log[k + 5].strip())
            xml_doc[frame][nelem].set('h', log[k + 6].strip())
            xml_doc[frame][nelem].set('angle', log[k + 7].strip())
            xml_doc[frame][nelem].set('pages', log[k + 8].strip())
            xml_doc[frame][nelem].set('group', log[k + 9].strip())
            xml_doc[frame][nelem].set('origratio', log[k + 10].strip())
            xml_doc[frame][nelem].set(
              'isorigratio', (1 if log[k + 11].strip() == '-1' else 0)
            )
            k += 11
          elif log[k][14:17] == 'arw':
            # Arrow
            xml_doc[frame][nelem].set('x1', log[k + 3].strip())
            xml_doc[frame][nelem].set('y1', log[k + 4].strip())
            xml_doc[frame][nelem].set('x2', log[k + 5].strip())
            xml_doc[frame][nelem].set('y2', log[k + 6].strip())
            xml_doc[frame][nelem].set('pages', log[k + 7].strip())
            xml_doc[frame][nelem].set('group', log[k + 8].strip())
            xml_doc[frame][nelem].set('lw', log[k + 9].strip())
            xml_doc[frame][nelem].set('ukoa', log[k + 10].strip()[1:])
            x1 = float(log[k + 3])
            y1 = float(log[k + 4])
            x2 = float(log[k + 5])
            y2 = float(log[k + 6])
            lw = float(log[k + 9])
            anglerad = -math.atan2((y2 - y1), (x2 - x1) * paperratio)
            angle = anglerad * 180 / math.pi
            w = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2 / paperratio**2)
            h = lw / paperh
            x = x1 - math.sin(anglerad) * h / paperratio / 2
            y = y1 - math.cos(anglerad) * h / 2
            xml_doc[frame][nelem].set('x', f'{x:.10f}')
            xml_doc[frame][nelem].set('y', f'{y:.10f}')
            xml_doc[frame][nelem].set('w', f'{w:.10f}')
            xml_doc[frame][nelem].set('h', f'{h:.10f}')
            xml_doc[frame][nelem].set('angle', f'{angle:.10f}')
            k += 10
          # Check if the element is in a group
          if xml_doc[frame][nelem].get('group') != '0':
            igroup = int(xml_doc[frame][nelem].get('group'))
            ngroups = max(int(xml_doc[frame].get('n_groups')), igroup)
            xml_doc[frame].set('n_groups', ngroups)
          nelem += 1
    k += 1


################################# Final parsing
def endparse(xml_doc):
  nframes = int(xml_doc.get('n_frames'))
  page = int(xml_doc.get('n_pages'))
  for k in range(nframes):
    if not xml_doc[k].get('first_page'):
      if k == 0:
        xml_doc[k].set('first_page', '1')
      else:
        xml_doc[k].set(
          'first_page',
          int(xml_doc[k - 1].get('first_page')) + int(xml_doc[k - 1].get('n_pages')),
        )
      if k + 1 == nframes:
        xml_doc[k].set('n_pages', page + 1 - int(xml_doc[k].get('first_page')))
      elif xml_doc[k + 1].get('first_page'):
        xml_doc[k].set(
          'n_pages',
          int(xml_doc[k + 1].get('first_page')) - int(xml_doc[k].get('first_page')),
        )
      else:
        xml_doc[k].set('n_pages', '1')
      xml_doc[k].set('n_elements', '0')


def parseall(jobname):
  """Parse tex + log of a document to construct the xml tree"""
  # Create root element of xml tree
  xml_doc = ET.Element('document')
  # Parse tex
  parsetex(jobname, xml_doc)
  # Parse log
  parselog(jobname, xml_doc)
  # Finalize the xml tree
  endparse(xml_doc)
  return xml_doc



"""
Copyright 2015-2024 Jean-Baptiste Delisle

This file is part of BEd.

BEd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BEd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with BEd.  If not, see <http://www.gnu.org/licenses/>.
"""

# This file defines the Document class (complete LaTeX-Beamer document)

import os

import pymupdf
from PySide6 import QtCore, QtGui

from . import parsing
from .editor import BedAdjustTextEdit, PropertiesEditor
from .frame import Frame
from .helpers import spacelines
from .settings import settings


class Document(QtCore.QObject):
  """Document class represents a complete LaTeX-Beamer document
  It contains a header (LaTeX preamble), Beamer frames, and a footer
  """

  def __init__(
    self,
    header='',
    footer='',
    frames=(),
    paper_w=None,
    paper_h=None,
    npages=None,
    modified=False,
  ):
    """Initialisation"""
    super().__init__()
    self.header = header
    self.footer = footer
    self.frames = []
    # Copy the frames instead of simply linking them
    for f in frames:
      self.frames.append(f.copy())
    self.nframes = len(self.frames)
    self.paper_w = paper_w
    self.paper_h = paper_h
    self.paper_ratio = paper_w / paper_h if (paper_w and paper_h) else None
    self.npages = npages
    self.modified = modified

  def copy(self):
    """Copy function (deep copy)"""
    return Document(
      self.header,
      self.footer,
      self.frames,
      self.paper_w,
      self.paper_h,
      self.npages,
      self.modified,
    )

  def readDoc(self, filename):
    """Read document from pdflatex compilation outputs"""
    imgs = self.loadpdf(filename + '.pdf')
    self.readxml(filename, imgs)

  def loadpdf(self, pdffile):
    """Extract images from the pdf file (output of pdflatex)"""
    pdf = pymupdf.open(pdffile)
    pixs = [page.get_pixmap(dpi=settings.dpi) for page in pdf]
    imgs = [
      QtGui.QImage(
        pix.samples,
        pix.width,
        pix.height,
        QtGui.QImage.Format_RGBA8888 if pix.alpha else QtGui.QImage.Format_RGB888,
      )
      for pix in pixs
    ]
    return imgs

  def readxml(self, filename, imgs):
    """Read the xml tree of the latex document"""
    xml_doc = parsing.parseall(filename)
    self.header = xml_doc.get('header').strip()
    self.footer = xml_doc.get('footer').strip()
    self.nframes = int(xml_doc.get('n_frames'))
    self.npages = int(xml_doc.get('n_pages'))
    self.paper_w = float(xml_doc.get('paper_w'))
    self.paper_h = float(xml_doc.get('paper_h'))
    self.paper_ratio = self.paper_w / self.paper_h
    self.modified = False
    self.frames = []
    for xml_frame in xml_doc:
      self.frames.append(Frame(xml_frame, imgs, self.paper_ratio, self.paper_h))

  def writeTeX(self, texfile, single_frame=-1):
    """Write the Document in a TeX file"""
    # Header
    tex = self.header + '\n'
    # Frames
    for kfr, fr in enumerate(self.frames):
      if single_frame in (-1, kfr):
        tex += fr.writeTeX() + '\n'
      elif fr.before.strip():
        # Put the in-between frames content anyway
        tex += spacelines(fr.before.strip(), 2) + '\n'
    # If compiling single frame, put a dummy frame at the end to capture
    # remaining in-between frames content (sections...)
    if single_frame > -1:
      tex += '\\begin{frame}\\end{frame}\n'
    # Footer
    tex += self.footer + '\n'
    # Writing
    with open(texfile, 'w') as f:
      f.write(tex)
    # Indent the tex file if indent_cmd available
    if settings.indent_cmd != '':
      if '%f' in settings.indent_cmd:
        os.system(settings.indent_cmd.replace('%f', texfile))
      else:
        os.system(settings.indent_cmd + ' ' + texfile)

  def edit(self, parentwindow):
    """Launch the PropertiesEditor (see editor.py) to edit the Document properties."""
    # Initial values of props (before editing)
    init_values = [self.header, self.footer]
    # Define BedWidget for each prop (see editor.py)
    bwheader = BedAdjustTextEdit(
      init_values[0], label=self.tr('Header'), highlight=True
    )
    bwfooter = BedAdjustTextEdit(
      init_values[1], label=self.tr('Footer'), highlight=True
    )
    bedwidgets = [bwheader, bwfooter]
    values = PropertiesEditor.getValues(
      parentwindow, bedwidgets, settings, 'document', self.tr('Document')
    )
    # Check if values have changed
    if values:
      # Update properties
      self.header = values[0]
      self.footer = values[1]
      return True
    return False

  def clean_imgs(self):
    """Clean all preview images from the document's frames."""
    for frame in self.frames:
      frame.clean_imgs()
      frame.clean_imgs()



"""
Copyright 2015-2024 Jean-Baptiste Delisle

This file is part of BEd.

BEd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BEd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with BEd.  If not, see <http://www.gnu.org/licenses/>.
"""

# This file defines the BEd properties editor (class PropertiesEditor).
# It is a simple QDialog with a FormLayout.
# Each line of the form is a property that can be edited (class BedWidget).
# A BedWidget is a QtWidget (QLineEdit, QSpinBox...) that has
# additionnal functions.
from PySide6 import QtCore, QtGui, QtWidgets

# Syntax highlighting (if available)
try:
  from .highlighter import BedHighlighter

  pygments_available = True
except ModuleNotFoundError:
  pygments_available = False


class AdjustTextEdit(QtWidgets.QTextEdit):
  """Class that inherits QTextEdit but with a minimum height adjusting to content"""

  def __init__(self, plainText='', parent=None):
    """Initialisation"""
    super().__init__(parent=parent)
    # Force the text to be plainText
    self.setAcceptRichText(False)
    self.setPlainText(plainText)

  def paintEvent(self, *args, **kwargs):
    """Painting"""
    # Get content size
    document_size = self.document().size().toSize()
    # Adjust minimum height to content
    self.setMinimumHeight(document_size.height() + 5)
    # Paint
    super().paintEvent(*args, **kwargs)


class ColorDialogButton(QtWidgets.QWidget):
  """Colored button that trigger a QColorDialog when clicked"""

  colorChanged = QtCore.Signal(QtGui.QColor, name='colorChanged')

  def __init__(self, color, parent=None):
    """Initialisation"""
    super().__init__(parent=parent)
    self.setMinimumSize(50, 25)
    self.setMaximumSize(50, 25)
    self.setColor(color)

  def setColor(self, color):
    """Change the button color and trigger colorChanged signal"""
    self.color = color
    self.repaint()
    self.colorChanged.emit(color)

  def paintEvent(self, *args, **kwargs):
    """Painting"""
    # Draw colored rectangle with gray border
    painter = QtGui.QPainter()
    painter.begin(self)
    painter.setBrush(self.color)
    painter.setPen(QtGui.QPen(QtGui.QColor('gray'), 1.0))
    painter.drawRect(QtCore.QRectF(0.5, 0.5, self.width() - 1, self.height() - 1))
    painter.end()

  def mouseReleaseEvent(self, event):
    """Launch QColorDialog when button is clicked"""
    color = QtWidgets.QColorDialog.getColor(initial=self.color, parent=self.parent())
    if color.isValid():
      self.setColor(color)


class BedWidget:
  """Class that contains one or several QtWidgets with additionnal functions"""

  def __init__(
    self, value, label=None, widgets=None, default=None, onChange=None, parent=None
  ):
    """Initialisation"""
    # Remember previous value
    self.prev_value = value
    # Label to be written in front of the widget in the form layout
    self.label = QtWidgets.QLabel(label) if label else None
    # Qt widgets embedded in the BedWidget
    self.widgets = widgets if widgets else []
    self.default = default
    if len(widgets) == 1 and default is None:
      self.container = widgets[0]
    else:
      self.container = QtWidgets.QWidget()
      layout = QtWidgets.QHBoxLayout(self.container)
      layout.setContentsMargins(0, 0, 0, 0)
      for widget in widgets:
        layout.addWidget(widget)
      if default is not None:
        self.resetbutton = QtWidgets.QPushButton(self.tr('Reset'))
        self.resetbutton.clicked.connect(self.resetValue)
        if value == default:
          self.resetbutton.setVisible(False)
        layout.addWidget(self.resetbutton)

    # Function to be called when the widget is modified
    self.onChange = onChange
    # Parent widget
    self.parent = parent

  def value(self):
    """Function that provide the value of the widget in a uniformed way"""
    return

  def setValue(self, value):
    """Function that set the value of the widget in a uniformed way"""

  def resetValue(self):
    self.setValue(self.default)
    self.resetbutton.setVisible(False)

  def onChangeConnect(self):
    """Connector that is raised when the widget is edited"""
    # Show/hide the reset button...
    if self.default is not None:
      if self.value() == self.default:
        self.resetbutton.setVisible(False)
      else:
        self.resetbutton.setVisible(True)
    # Call the onChange function if it is defined and needed
    if self.onChange and not self.parent.changing:
      self.onChange(self.parent)
    # Remember previous value
    self.prev_value = self.value()

  def tr(self, sourceText, disambiguation=''):
    return QtCore.QCoreApplication.translate('BedWidget', sourceText, disambiguation)


class BedLineEdit(BedWidget):
  """BedWidget containing a QLineEdit"""

  def __init__(
    self, value, label=None, default=None, onChange=None, highlight=False, parent=None
  ):
    """Initialisation"""
    super().__init__(
      value, label, [QtWidgets.QLineEdit(value)], default, onChange, parent
    )
    self.widgets[0].editingFinished.connect(self.onChangeConnect)
    if highlight and pygments_available:
      self.highlighter = BedHighlighter(self.widgets[0].document())

  def value(self):
    return self.widgets[0].text()

  def setValue(self, value):
    self.widgets[0].setText(value)
    self.prev_value = self.value()


class BedAdjustTextEdit(BedWidget):
  """BedWidget containing an AdjustTextEdit"""

  def __init__(
    self, value, label=None, default=None, onChange=None, highlight=False, parent=None
  ):
    """Initialisation"""
    super().__init__(
      value, label, [AdjustTextEdit(value, parent)], default, onChange, parent
    )
    self.widgets[0].textChanged.connect(self.onChangeConnect)
    if highlight and pygments_available:
      self.highlighter = BedHighlighter(self.widgets[0].document())

  def value(self):
    return self.widgets[0].toPlainText()

  def setValue(self, value):
    self.widgets[0].setPlainText(value)
    self.prev_value = self.value()


class BedSpinBox(BedWidget):
  """BedWidget containing a QSpinBox"""

  def __init__(
    self, value, mini, maxi, step, label=None, default=None, onChange=None, parent=None
  ):
    """Initialisation"""
    super().__init__(value, label, [QtWidgets.QSpinBox()], default, onChange, parent)
    self.widgets[0].setRange(mini, maxi)
    self.widgets[0].setSingleStep(step)
    self.widgets[0].setValue(value)
    self.widgets[0].editingFinished.connect(self.onChangeConnect)

  def value(self):
    return self.widgets[0].value()

  def setValue(self, value):
    self.widgets[0].setValue(value)
    self.prev_value = self.value()


class BedDoubleSpinBox(BedWidget):
  """BedWidget containing a QDoubleSpinBox"""

  def __init__(
    self,
    value,
    mini,
    maxi,
    step,
    decimals,
    label=None,
    default=None,
    onChange=None,
    parent=None,
  ):
    """Initialisation"""
    super().__init__(
      value, label, [QtWidgets.QDoubleSpinBox()], default, onChange, parent
    )
    self.widgets[0].setRange(mini, maxi)
    self.widgets[0].setSingleStep(step)
    self.widgets[0].setDecimals(decimals)
    self.widgets[0].setValue(value)
    self.widgets[0].editingFinished.connect(self.onChangeConnect)

  def value(self):
    return self.widgets[0].value()

  def setValue(self, value):
    self.widgets[0].setValue(value)
    self.prev_value = self.value()


class BedCheckBox(BedWidget):
  """BedWidget containing a QCheckBox"""

  def __init__(
    self, value, label=None, checklabel=None, default=None, onChange=None, parent=None
  ):
    """Initialisation"""
    super().__init__(
      value, label, [QtWidgets.QCheckBox(checklabel)], default, onChange, parent
    )
    self.widgets[0].setChecked(value)
    self.widgets[0].stateChanged.connect(self.onChangeConnect)

  def value(self):
    return self.widgets[0].checkState() == QtCore.Qt.Checked

  def setValue(self, value):
    self.widgets[0].setChecked(value)
    self.prev_value = self.value()


class BedComboBox(BedWidget):
  """BedWidget containing a QComboBox"""

  def __init__(
    self,
    value,
    listvals,
    label=None,
    checklabel=None,
    default=None,
    onChange=None,
    parent=None,
  ):
    """Initialisation"""
    super().__init__(value, label, [QtWidgets.QComboBox()], default, onChange, parent)
    self.widget = QtWidgets.QComboBox()
    self.widgets[0].addItems(listvals)
    self.widgets[0].setCurrentIndex(value)
    self.widgets[0].currentIndexChanged.connect(self.onChangeConnect)

  def value(self):
    return self.widgets[0].currentIndex()

  def setValue(self, value):
    self.widgets[0].setCurrentIndex(value)
    self.prev_value = self.value()


class BedColorPicker(BedWidget):
  """BedWidget containing a ColorDialogButton"""

  def __init__(self, value, label=None, default=None, onChange=None, parent=None):
    """Initialisation"""
    super().__init__(
      value, label, [ColorDialogButton(value, parent)], default, onChange, parent
    )
    self.widgets[0].colorChanged.connect(self.onChangeConnect)

  def value(self):
    return self.widgets[0].color

  def setValue(self, value):
    self.widgets[0].setColor(value)
    self.prev_value = self.value()


class BedPenPicker(BedWidget):
  """BedWidget containing a ColorDialogButton + QDoubleSpinBox"""

  def __init__(self, value, label=None, default=None, onChange=None, parent=None):
    """Initialisation"""
    super().__init__(
      value,
      label,
      [ColorDialogButton(value.color(), parent), QtWidgets.QDoubleSpinBox()],
      default,
      onChange,
      parent,
    )
    self.widgets[0].colorChanged.connect(self.onChangeConnect)
    self.widgets[1].setRange(0, 100)
    self.widgets[1].setSingleStep(0.1)
    self.widgets[1].setDecimals(2)
    self.widgets[1].setValue(value.widthF())
    self.widgets[1].editingFinished.connect(self.onChangeConnect)

  def value(self):
    return QtGui.QPen(self.widgets[0].color, self.widgets[1].value())

  def setValue(self, value):
    self.widgets[0].setColor(value.color())
    self.widgets[1].setValue(value.widthF())
    self.prev_value = self.value()


class BedPushButton(BedWidget):
  """BedWidget containing a QPushButton"""

  def __init__(self, label=None, buttonlabel=None, onChange=None, parent=None):
    """Initialisation"""
    super().__init__(
      None, label, [QtWidgets.QPushButton(buttonlabel)], None, onChange, parent
    )
    self.widgets[0].clicked.connect(self.onChangeConnect)


class PropertiesEditor(QtWidgets.QDialog):
  """Dialog that ask the user for changes in objects properties"""

  def __init__(self, parent, bedwidgets, settings, objtype, objname):
    """Initialisation"""
    super().__init__(parent)
    self.setWindowTitle(self.tr('BEd - Properties Editor') + ' (' + objname + ')')
    self.setMinimumSize(300, 300)
    self.resize(
      settings.__dict__['editor_' + objtype + '_width'],
      settings.__dict__['editor_' + objtype + '_height'],
    )
    self.bedwidgets = bedwidgets
    self.changing = False
    self.settings = settings
    self.objtype = objtype
    globlayout = QtWidgets.QVBoxLayout(self)
    # Scrollbar around form
    scroll = QtWidgets.QScrollArea(self)
    globlayout.addWidget(scroll)
    scroll.setWidgetResizable(True)
    view = QtWidgets.QWidget(scroll)
    scroll.setWidget(view)
    # Formlayout listing all properties
    proplayout = QtWidgets.QFormLayout(view)
    proplayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.ExpandingFieldsGrow)
    view.setLayout(proplayout)
    # Loop over properties (BedWidgets)
    for bwid in bedwidgets:
      bwid.parent = self
      # Put a name in front of the widgets container if bwid.label is defined
      if bwid.label:
        proplayout.addRow(bwid.label, bwid.container)
      else:
        proplayout.addRow(bwid.container)

    # OK and Cancel buttons
    buttons = QtWidgets.QDialogButtonBox(
      QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
      QtCore.Qt.Horizontal,
      self,
    )
    globlayout.addWidget(buttons)

    # Connectors
    buttons.accepted.connect(self.accept)
    buttons.rejected.connect(self.reject)

  @staticmethod
  def getValues(parent, bedwidgets, settings, objtype, objname):
    """Static method that is called in the main file to get
    the final values of the paramaters (if they changed)
    """
    QtWidgets.QApplication.restoreOverrideCursor()
    dialog = PropertiesEditor(parent, bedwidgets, settings, objtype, objname)
    ok = dialog.exec_()
    if ok:
      # Return the list of parameters
      return [bwid.value() for bwid in dialog.bedwidgets]
    # Return None if the user canceled
    return None

  def resizeEvent(self, event):
    """Things to do when the user resize the window"""
    # Most important resizing consequences are treated by the painterArea resizeEvent
    # Here we just save the new size in the application settings
    self.settings.__dict__['editor_' + self.objtype + '_width'] = self.width()
    self.settings.__dict__['editor_' + self.objtype + '_height'] = self.height()



"""
Copyright 2015-2024 Jean-Baptiste Delisle

This file is part of BEd.

BEd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BEd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with BEd.  If not, see <http://www.gnu.org/licenses/>.
"""

# This file defines the PainterArea class,
# which deals with the drawings
# and the mouse events

import time

from PySide6 import QtCore, QtGui, QtWidgets

from .element import Arrow
from .settings import settings


class Painter(QtGui.QPainter):
  """QPainter customization"""

  def __init__(self, parent=None):
    """Initialisation"""
    super().__init__()
    self.parent = parent

  def drawVline(self, x):
    """Draw a vertical line at x"""
    self.drawLine(
      QtCore.QLineF(
        x * self.parent.width(), 0, x * self.parent.width(), self.parent.height()
      )
    )

  def drawHline(self, y):
    """Draw an horizontal line at y"""
    self.drawLine(
      QtCore.QLineF(
        0, y * self.parent.height(), self.parent.width(), y * self.parent.height()
      )
    )


class PainterScrollArea(QtWidgets.QScrollArea):
  """Painter scrolling Area: QScrollArea around the painterArea."""

  def __init__(self, parent, painterArea):
    """Initialisation"""
    super().__init__(parent)
    self.setAlignment(QtCore.Qt.AlignCenter)
    self.setWidgetResizable(True)
    self.setWidget(painterArea)
    self.painterArea = painterArea

  def wheelEvent(self, event):
    """Wheel event: check if the painterArea intercepted it"""
    if not self.painterArea.on_wheelEvent(event):
      super().wheelEvent(event)
      self.painterArea.lastx = event.position().x() - self.painterArea.pos().x()
      self.painterArea.lasty = event.position().y() - self.painterArea.pos().y()
      self.painterArea.refreshCursor()

  def resizeEvent(self, event):
    """resizeEvent: Apply to self then to painterArea"""
    super().resizeEvent(event)
    self.painterArea.resizeEvent(event)


class PainterArea(QtWidgets.QWidget):
  """Painter Area, where everything is drawn"""

  def __init__(self, parent=None):
    """Initialisation"""
    super().__init__(parent)
    self.setFocus()
    self.setMouseTracking(True)
    self.painter = Painter(self)
    # link to bed main window
    self.parent = parent
    # init of mouse
    self.mousepressed = False
    self.lastx = -1
    self.lasty = -1
    # init of keyboard modifiers
    self.controlpressed = False
    # init of temporary group
    # (group that is create when moving/resizing... a multi-selection)
    self.tmpgroup = False
    # init of resize object flag
    self.resizeObj = None
    # init of mouse wheel flag
    self.wheelObj = None
    # init of rectangular selection flag
    self.rectSelectionOn = False
    # Init of painting area last refresh time
    self.last_refresh = time.time()
    # Init of zooming status
    self.zooming = False
    # Init of scrollArea
    self.scrollArea = PainterScrollArea(parent, self)

  def setSize(self):
    """Compute where to draw the frame given the window size and the frame w/h ratio."""
    sAsize = self.scrollArea.maximumViewportSize()
    w = min(sAsize.width(), round(sAsize.height() * self.parent.document.paper_ratio))
    h = round(w / self.parent.document.paper_ratio)
    if (not self.zooming) or (self.width() <= w and self.height() <= h):
      self.zooming = False
      self.setFixedSize(w, h)

  def resizeEvent(self, event):
    """Recompute the geometry if window is resized"""
    if self.parent.frame is not None:
      self.setSize()

  def paintEvent(self, event):
    """Drawing frame..."""
    self.painter.begin(self)
    if self.parent.isCompiling:
      font = self.painter.font()
      font.setPointSize(20)
      self.painter.setFont(font)
      rect = QtCore.QRect(0, 0, self.width(), self.height())
      self.painter.drawText(
        rect, QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter, self.tr('Compiling...')
      )
    elif self.parent.frame is not None:
      # Ask the frame to draw itself at correct position
      self.parent.document.frames[self.parent.frame].draw(
        self.parent.framePage, self.painter
      )
      # Draw the mouse rectangular selection
      if self.rectSelectionOn:
        self.painter.setPen(settings.pen_mouse_selection)
        self.painter.drawRect(self.rectSelectionRect)
    self.painter.end()

  def mousePressEvent(self, event):
    """What to do when mouse is pressed"""
    if (
      QtWidgets.QApplication.overrideCursor()
      and QtWidgets.QApplication.overrideCursor().shape() == QtCore.Qt.OpenHandCursor
    ):
      QtWidgets.QApplication.restoreOverrideCursor()
      QtWidgets.QApplication.setOverrideCursor(
        QtGui.QCursor(QtCore.Qt.ClosedHandCursor)
      )
    if self.parent.frame is not None:
      self.mousepressed = True
      # Reset mouse-moved test
      self.mousemoved = False
      # Unselect wheel object (see wheelEvent)
      self.wheelObj = None
      # Save initial position...
      self.mousepressx = event.position().x()
      self.mousepressy = event.position().y()
      self.lastx = event.position().x()
      self.lasty = event.position().y()
      # If control is not pressed, all previously selected elements are unselected
      if not (
        QtWidgets.QApplication.queryKeyboardModifiers() & QtCore.Qt.ControlModifier
      ):
        # If control is not pressed, unselect all elements
        self.controlpressed = False
        for obj in (
          self.parent.document.frames[self.parent.frame].elements
          + self.parent.document.frames[self.parent.frame].groups
        ):
          obj.selected = False
        self.repaint()
      else:
        self.controlpressed = True

      # Look if an element is under the mouse
      # Reverse the list to prefere to select foreground elements
      for el in reversed(self.parent.document.frames[self.parent.frame].elements):
        obj = el.group if el.group else el
        if obj.hover > 0:
          # The element is under the mouse
          # Toggle selection
          obj.selected = not obj.selected
          self.repaint()
          # Stop to search if something is selected
          return

  def mouseDoubleClickEvent(self, event):
    # Double click -> edit the selected object (or the frame if no selection)
    self.parent.editAny()

  def mouseMoveEvent(self, event):
    """What to do if mouse is moved"""
    if self.parent.frame is not None:
      # If the mouse is pressed
      # There is 2 possibilities
      # 1) we are resizing/moving something
      # 2) we are drawing a rectangular mouse selection
      if self.mousepressed:
        # First time mouse is moved
        if not self.mousemoved:
          if self.controlpressed:
            # Control pressed:
            # There can be 0,1 or more selected elements + the one under mouse:
            hover = 0
            # Look if something under mouse
            for el in reversed(self.parent.document.frames[self.parent.frame].elements):
              obj = el.group if el.group else el
              if obj.hover > 0:
                # Make sure to select the element is under the mouse
                obj.selected = True
                hover = obj.hover if obj.hover < 1100 else 1000
                break
            # If something is found: initialise moving/resizing of the group of selected
            # elements
            if hover > 0:
              # Undo snapshot before doing things
              self.parent.undoStack.doTemp()
              # Save frame groups list
              # Create group
              self.parent.document.frames[self.parent.frame].tmpgroup(
                self.parent.document.paper_ratio
              )
              # Use it as resizing object
              self.resizeObj = self.parent.document.frames[self.parent.frame].groups[-1]
              # If only 1 element in group use the element instead
              if len(self.resizeObj.elements) == 1:
                self.resizeObj = self.resizeObj.elements[0]
                self.resizeObj.refreshHover(
                  self.mousepressx / self.width(), self.mousepressy / self.height()
                )
                self.parent.document.frames[self.parent.frame].untmpgroup()
              else:
                # Remember to delete the group when mouse released
                self.tmpgroup = True
                self.resizeObj.hover = hover
              self.resizeObj.saveGeometry()
          else:
            # Control not pressed: 0 or 1 selected element
            for el in reversed(self.parent.document.frames[self.parent.frame].elements):
              obj = el.group if el.group else el
              if obj.selected:
                # The element is selected
                # Initialise resizing/moving of element
                self.parent.undoStack.doTemp()
                self.resizeObj = obj
                obj.saveGeometry()
        # Remember that mouse already moved
        self.mousemoved = True
        # To avoid to waste computing time
        # Do not refresh to much
        tm = time.time()
        if tm - self.last_refresh > settings.refresh_interval:
          self.last_refresh = tm
          if self.resizeObj:
            # Case 1)
            Dx = (event.position().x() - self.mousepressx) / self.width()
            Dy = (event.position().y() - self.mousepressy) / self.height()
            dx = (event.position().x() - self.lastx) / self.width()
            dy = (event.position().y() - self.lasty) / self.height()
            self.lastx = event.position().x()
            self.lasty = event.position().y()
            # ask the object to resize itself
            self.resizeObj.change_geom(
              Dx,
              Dy,
              dx,
              dy,
              self.parent.document.frames[self.parent.frame].xGuides(
                self.parent.framePage
              ),
              self.parent.document.frames[self.parent.frame].yGuides(
                self.parent.framePage
              ),
            )
          else:
            # Case 2)
            # Draw mouse selection
            self.rectSelectionOn = True
            self.rectSelection(event.position().x(), event.position().y())
          self.repaint()
      else:
        # The mouse is not pressed
        # We look if something is under the mouse to change the shape of the cursor
        # (resizing, moving cursors...)
        self.lastx = event.position().x()
        self.lasty = event.position().y()
        self.refreshCursor()

  def on_wheelEvent(self, event):
    """Wheel event : call wheel function of obj if one is selected"""
    if self.parent.frame is not None:
      if QtWidgets.QApplication.queryKeyboardModifiers() & QtCore.Qt.ControlModifier:
        self.zooming = True
        w = round(self.width() * 1.1 ** (event.angleDelta().y() / 120))
        h = round(w / self.parent.document.paper_ratio)
        x = round(
          event.position().x()
          + (self.pos().x() - event.position().x()) * w / self.width()
        )
        y = round(
          event.position().y()
          + (self.pos().y() - event.position().y()) * h / self.height()
        )
        self.setFixedSize(w, h)
        self.scrollArea.horizontalScrollBar().setValue(-x)
        self.scrollArea.verticalScrollBar().setValue(-y)
        self.lastx = event.position().x() - self.pos().x()
        self.lasty = event.position().y() - self.pos().y()
        self.refreshCursor()
        return True
      for el in reversed(self.parent.document.frames[self.parent.frame].elements):
        # Only arrows have a wheel function (for now)
        if el.selected and type(el) is Arrow and el.hover == 1000:
          # Check if object has changed
          if el != self.wheelObj:
            # Save state before wheel event
            self.parent.undoStack.do()
            # Save which object is selected
            self.wheelObj = el
            el.saveGeometry()
          # object wheel event
          el.wheel(event.angleDelta().x(), event.angleDelta().y())
          self.repaint()
          return True
    return False

  def mouseReleaseEvent(self, event):
    """What to do when mouse is released"""
    # Refresh cursor
    self.refreshCursor()
    if self.parent.frame is not None:
      # If a resizing was initiate but the mouse did not move
      # cancel the undo stack temp (see undo.py)
      if self.resizeObj:
        if self.mousemoved:
          self.parent.undoStack.confirmTemp()
        else:
          self.parent.undoStack.cancelTemp()
      # Re-init variables
      self.mousepressed = False
      self.rectSelectionOn = False
      self.resizeObj = None
      if self.tmpgroup:
        self.parent.document.frames[self.parent.frame].untmpgroup()
        self.tmpgroup = False
      self.repaint()

  def keyPressEvent(self, event):
    """What to do when a key is pressed"""
    if not self.mousepressed:
      self.refreshCursor()  # In case modifiers keys have changed

  def keyReleaseEvent(self, event):
    """What to do when a key is pressed"""
    if not self.mousepressed:
      self.refreshCursor()  # In case modifiers keys have changed

  def refreshCursor(self):
    """Check what is under the mouse to change the shape of the cursor
    (resizing, moving, rotating...)
    """
    if self.parent.frame is not None:
      x = self.lastx / self.width()
      y = self.lasty / self.height()
      hover = 0
      cursor = None
      # Reverse the list to prefere foreground elements
      for el in reversed(self.parent.document.frames[self.parent.frame].elements):
        obj = el.group if el.group else el
        if obj.isvisible(self.parent.framePage):
          hover = obj.refreshHover(x, y)
          if hover > 0:
            # Set cursor shape accordingly
            cursor = obj.selectCursor()
            # Update cursor only if needed
            if not QtWidgets.QApplication.overrideCursor():
              QtWidgets.QApplication.setOverrideCursor(QtGui.QCursor(cursor))
            elif QtWidgets.QApplication.overrideCursor().shape() != cursor:
              QtWidgets.QApplication.restoreOverrideCursor()
              QtWidgets.QApplication.setOverrideCursor(QtGui.QCursor(cursor))
            # Stop searching if something is found
            break
      # Restore cursor if needed
      if cursor is None and QtWidgets.QApplication.overrideCursor():
        QtWidgets.QApplication.restoreOverrideCursor()

  def rectSelection(self, x, y):
    """Function that decide which elements/groups to select
    when doing a rectangular mouse selection
    """
    # The criterium is that more than half of the element surface
    # must be inside the mouse rectangle
    # Mouse rectangle
    # In painterArea coordinates
    left = min(x, self.mousepressx)
    right = max(x, self.mousepressx)
    top = min(y, self.mousepressy)
    bottom = max(y, self.mousepressy)
    self.rectSelectionRect = QtCore.QRect(left, top, right - left, bottom - top)

    # In frame coordinates
    left = left / self.width()
    right = right / self.width()
    top = top / self.height()
    bottom = bottom / self.height()

    # Find selected objects
    for obj in (
      self.parent.document.frames[self.parent.frame].elements
      + self.parent.document.frames[self.parent.frame].groups
    ):
      if obj.group is None:
        # Object rectangle
        oL = obj.get_xguide('left')
        oR = max(obj.get_xguide('right'), oL + settings.epsilon)  # in case width=0
        oT = obj.get_yguide('top')
        oB = max(obj.get_yguide('bottom'), oT + settings.epsilon)  # in case height=0
        oS = (oR - oL) * (oB - oT)
        # Intersection rectangle
        cL = max(left, oL)
        cR = min(right, oR)
        cT = max(top, oT)
        cB = min(bottom, oB)
        # Intersection surface
        cS = max(0, cR - cL) * max(0, cB - cT)
        # Criterium
        obj.selected = cS > oS / 2 and obj.isvisible(self.parent.framePage)

  def selectAll(self):
    """Select all elements and groups"""
    if self.parent.frame is not None:
      for obj in (
        self.parent.document.frames[self.parent.frame].elements
        + self.parent.document.frames[self.parent.frame].groups
      ):
        # Do not select elements that are contained in a group (just select the group)
        if obj.group is None and obj.isvisible(self.parent.framePage):
          obj.selected = True
      self.repaint()

  def unselectAll(self):
    """Unselect all elements and groups"""
    if self.parent.frame is not None:
      for obj in (
        self.parent.document.frames[self.parent.frame].elements
        + self.parent.document.frames[self.parent.frame].groups
      ):
        if obj.group is None:
          obj.selected = False
      self.repaint()
